/******************************************************************************/
/*                                                                            */
/*                               audioseg.h                                   */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Guig                                                       Sun Aug  4 2002 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2003 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 * $Author: ggravier $
 * $Date: 2003/11/20 15:50:57 $
 * $Revision: 1.11 $
 *
 */

/*
 * Audio Segmentation Library Header.
 */

#ifndef _audioseg_h_
# define _audioseg_h_

# include <system.h>
# include <spro.h>

     /* ---------------------------- */
     /* ----- some hard limits ----- */
     /* ---------------------------- */

# define ASEG_NULL_TIME  -1.0          /* null time                            */

# define GMM_NUM_COMP_MAX   4096       /* maximum number of components         */
# define GMM_MIN_WEIGHT     1e-5       /* minimum component weight             */

# define W_UP 1                        /* update weights                       */
# define M_UP 2                        /* update means                         */
# define V_UP 4                        /* update variances                     */

     /* ---------------------------- */
     /* ----- type definitions ----- */
     /* ---------------------------- */

typedef struct {
  unsigned short nlabels;       /* number of labels                           */
  char **name;                  /* segment labels                             */
} seglab_t;  

typedef struct asseg_s {
  seglab_t *label;              /* segment label                              */
  float st, et;                 /* segment start and end times                */
  float score;                  /* segment score (whatever this is)           */
  struct asseg_s *next;         /* next segment in a "transcription"          */
  struct asseg_s *prev;         /* next segment in a "transcription"          */
} asseg_t;

typedef struct {
  unsigned short n;             /* number of Gaussian components              */
  unsigned short dim;           /* dimension                                  */
  float *w;                     /* component weights                          */
  float **m;                    /* mean vectors                               */
  float **v;                    /* (inverse) diagonal covariance vectors      */
  double *logc;                 /* some constants                             */
  double *expz;
} gmm_t;                        /* Gaussian mixture model                     */

typedef struct {
  unsigned short n;             /* number of Gaussian components              */
  unsigned short dim;           /* dimension                                  */
  double *w;                    /* component weight accumulator               */
  double **m;                   /* array of mean accumulators                 */
  double **v;                   /* array of covariance accumulators           */
} gmmacc_t;                     /* accumulators                               */

     /* ---------------------------------------- */
     /* ----- segment allocation functions ----- */
     /* ---------------------------------------- */

/* allocate a segment */
asseg_t *seg_alloc(void);

/* create a segment (with attributes) */
asseg_t *seg_create(
  char *,                       /* labels separated by '+'                    */
  float,                        /* start time                                 */
  float,                        /* end time                                   */
  float                         /* score                                      */
);

/* free a segment */
void seg_free(
  asseg_t *                     /* segment address                            */
);

/* free a segment list */
void seg_list_free(
  asseg_t *                     /* pointer to the first segment               */
);

/* allocate memory for a segment label */
seglab_t *seg_label_alloc(
  unsigned short                /* number of labels                           */
);

/* free a segment label */
void seg_label_free(
  seglab_t *                    /* label to free                              */
);

     /* ------------------------------------ */
     /* ----- label handling functions ----- */
     /* ------------------------------------ */

/* compare a two labels, comparison is always true if one of the
   label is NULL. */
int seg_label_compare(
  seglab_t *,                   /* label                                      */
  seglab_t *                    /* label                                      */
);

/* return index of name in the label name array, or -1 if not
   found. */
int seg_label_name_index(
  seglab_t *,                   /* label                                      */
  const char *                  /* name                                       */
);

/* label string to table */
seglab_t *str_to_label(
  const char *                  /* label bames separated by '+'               */
);

# define get_num_labels(p)            ((p)->nlabels)
# define get_label_name(p, i)         ((p)->name[i])

     /* --------------------------------- */
     /* ----- segment I/O functions ----- */
     /* --------------------------------- */

/* read a segmentation */
asseg_t *seg_read(
  const char *                  /* input stream name (or NULL for stdin)      */
);

/* write a segmentation */
int seg_write(
  asseg_t *,                    /* pointer to the first segment               */
  const char *,                 /* output stream name (or NULL for stdout)    */
  const char *                  /* format string (lsep)                       */
);

     /* --------------------------------------- */
     /* ----- segment attribute accessors ----- */
     /* --------------------------------------- */

  
/* set segment labels, return the number of labels set or -1 in ase of
   error */
int set_seg_label(
  asseg_t *,                    /* pointer to the first segment               */
  char *                        /* labels separated by '+'                    */
);

/* add labels to the segment, return the number of labels set or -1 in
   case of error */
int add_seg_label(
  asseg_t *,                    /* pointer to the first segment               */
  char *                        /* labels separated by '+'                    */
);

# define get_seg_label(s)             ((s)->label)
# define get_seg_num_labels(s)        (get_num_labels(get_seg_label(s)))
# define get_seg_label_name(s, i)     (get_label_name(get_seg_label(s), i))
# define get_seg_start_time(s)        ((s)->st)
# define get_seg_end_time(s)          ((s)->et)
# define get_seg_score(s)             ((s)->score)
# define get_seg_next(s)              ((s)->next)
# define get_seg_prev(s)              ((s)->prev)

# define set_seg_start_time(s, a)     ((s)->st = (a))
# define set_seg_end_time(s, a)       ((s)->et = (a))
# define set_seg_score(s, a)          ((s)->score = (a))
# define set_seg_next(s, p)           ((s)->next = (p))
# define set_seg_prev(s, p)           ((s)->prev = (p))

     /* -------------------------------------------- */
     /* ----- likelihood computation functions ----- */
     /* -------------------------------------------- */

/* pre-compute constants for a model */
void gmm_const_set(
  gmm_t *                       /* model                                      */
);

/* compute segment log-likelihood */
double gmm_segment_log_like(
  spfstream_t *,                /* input stream                               */	
  unsigned long *,              /* segment length (0 to end of stream)        */  
  gmm_t *                       /* model                                      */
);

/* computes the log-likelihood of a frame for a gaussian mixture model */
double gmm_frame_log_like(
  spf_t *,                      /* feature vector                             */
  gmm_t *                       /* model                                      */
);

/* computes the log-likelihood of a frame for a gaussian mixture model */
double gmm_nbest_frame_log_like(
  spf_t *,                      /* feature vector                             */
  gmm_t *,                      /* model                                      */
  int *,                        /* indexes of Gaussian's to score             */
  unsigned short,               /* number of Gaussians to score               */
  int                           /* find out N-best or use existing indexes    */
);

     /* ----------------------------- */
     /* ----- GMM I/O functions ----- */
     /* ----------------------------- */

/* allocate memory for a model */
gmm_t *gmm_alloc(
  unsigned short,               /* number of Gaussian components             */
  unsigned short                /* dimension                                 */
);

/* free memory */
void gmm_free(
  gmm_t *                       /* model                                     */
);

/* read model from file */
gmm_t *gmm_read(
  const char *                  /* input filename                            */
);

/* write model to file */
int gmm_write(
  const char *,                 /* output filename                           */
  gmm_t *                       /* model                                     */
);

/* display in ASCII format -- should be in sgmcopy.c but I keep it
   here for debug purposes for the moment. */
void gmm_print(
  FILE *,                       /* output stream                             */
  gmm_t *                       /* model                                     */
);

     /* ---------------------------------------------- */
     /* ----- GMM parameter estimation functions ----- */
     /* ---------------------------------------------- */

/* allocate accumulators */
gmmacc_t *gmm_acc_alloc(
  unsigned short,               /* number of Gaussian components             */
  unsigned short                /* dimension                                 */
);

/* free accumulators */
void gmm_acc_free(
  gmmacc_t *                    /* accumulator                               */
);

/* Reset accumulator's values to zero */
void gmm_acc_reset(
  gmmacc_t *                    /* accumulator                               */
);

/* compute Gaussian occupation probabilities and return the frame log-prob. */
double gmm_occ_probs(
  gmm_t *,                      /* model                                     */
  spf_t *,                      /* feature vector                            */
  double *                      /* output occupation probabilities           */
);

/* sum up statistics in the accumulators for a segment */
double gmm_accumulate(
  spfstream_t *,                /* input feature stream                      */
  unsigned long *,              /* number of frames (0 to end of stream)     */
  gmm_t *,                      /* model                                     */
  gmmacc_t *                    /* accumulators                              */
);

/* ML re-estimation */
void gmm_ml_update(
  gmm_t *,                      /* updated model                             */
  gmmacc_t *,                   /* accumulators                              */
  unsigned long,                /* number of accumulated frames              */
  int,                          /* update flag                               */
  double,                       /* weights floor                             */
  double *                      /* variance floor                            */
);

/* MAP re-estimation */
void gmm_map_update(
  gmm_t *,                      /* updated model                             */
  gmmacc_t *,                   /* accumulators                              */
  unsigned long,                /* number of accumulated frames              */
  int,                          /* update flag                               */
  double,                       /* weights floor                             */
  double *,                     /* variance floor                            */
  double,                       /* prior weight                              */
  gmm_t *                       /* prior model                               */
);

/* set variance floor vector */
double *set_var_floor(
  float,                        /* asbolute floor or variance fraction       */ 
  const char *,                 /* global variance filename or NULL          */
  unsigned short                /* feature vector dimension                  */
);

     /* --------------------------------------- */
     /* ----- segment attribute accessors ----- */
     /* --------------------------------------- */

# define get_gmm_num_comp(g)          ((g)->n)
# define get_gmm_dim(g)               ((g)->dim)
# define get_gmm_weights(g)           ((g)->w)

# define get_gmm_comp_weight(g, i)    (*((g)->w+i))
# define get_gmm_comp_mean(g, i)      (*((g)->m+i))
# define get_gmm_comp_ivar(g, i)      (*((g)->v+i))
# define get_gmm_comp_logc(g, i)      (*((g)->logc+i))
# define get_gmm_comp_expz(g, i)      (*((g)->expz+i))

# define set_gmm_comp_weight(g, i, a) (*((g)->w+i) = (a))
# define set_gmm_comp_logc(g, i, a)   (*((g)->logc+i) = (a))
# define set_gmm_comp_expz(g, i, a)   (*((g)->expz+i) = (a))

# define set_gmm_num_comp(g, a)       ((g)->n = (a))
# define set_gmm_dim(g, a)            ((g)->dim = (a))

# define get_acc_num_comp(g)          ((g)->n)
# define get_acc_dim(g)               ((g)->dim)
# define get_acc_weights(g)           ((g)->w)
# define get_acc_comp_weight(g, i)    (*((g)->w+i))
# define get_acc_comp_mean(g, i)      (*((g)->m+i))
# define get_acc_comp_var(g, i)       (*((g)->v+i))

# ifdef _gmm_c_
#  define LOG_2_PI ((double)1.8378770664093453390819377091247588396072) /* ln(2 * M_PI) */
void swap_bytes(void *, unsigned long, size_t);
# endif

#endif /* _audioseg_h_ */
