/******************************************************************************/
/*                                                                            */
/*                                  bic.c                                     */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Guig                                                             July 2003 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2003 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 *   $Author: ggravier $
 *   $Date: 2003/08/25 13:31:18 $
 *   $Revision: 1.6 $
 *
 */

/*
 * BIC segmentation utilities.
 */

#define _bic_c_

#include <bic.h>

/* ----------------------------------------------------------------------------------------------- */
/* ----- bicwin_t *bic_init(spfstream_t *, unsigned long, unsigned long, unsigned long, int) ----- */
/* ----------------------------------------------------------------------------------------------- */
/*
 * Initialize the BIC window, i.e. allocate memory, set the window
 * stream and max_index (if any) and fill in the buffer.
 */
bicwin_t *bic_init(spfstream_t *f, unsigned long start, unsigned long size, unsigned long end, int fullc)
{
  bicwin_t *p;
  unsigned short dim = spf_stream_dim(f);

  /* allocate memory */
  if ((p = (bicwin_t *)malloc(sizeof(bicwin_t))) == NULL) {
    fprintf(stderr, "bic_init() -- cannot allocate memory\n");
    return(NULL);
  }

  p->f = f;
  p->buf = 0;
  p->maxi = end;
  p->offset = 0;
  p->is = 0;
  p->im = 0;
  p->ie = 0;
  if (fullc)
    p->cp = dim + 0.5 * dim * (dim + 1);
  else
    p->cp = 2 * dim;

  if ((p->buf = spf_buf_alloc(dim, size * dim * sizeof(float))) == NULL) {
    fprintf(stderr, "bic_init() -- cannot allocate memory\n");
    free(p);
    return(NULL);
  }
  
  if (bic_win_fill(p, start, size) == 0) {
    fprintf(stderr, "bic_init() -- cannot fill in BIC buffer\n");
    bic_reset(p);
    return(NULL);
  }

#ifdef HAVE_GSL
  if (fullc) {

    /* allocate for memory for matrix inversion with GSL */
    gsl_set_error_handler_off();
    if ((_S = gsl_matrix_alloc(dim, dim)) == NULL) {
      fprintf(stderr, "bic_init() -- cannot allocate memory\n");
      bic_reset(p);
      return(NULL);
    }

    if ((_m = gsl_vector_alloc(dim)) == NULL) {
      fprintf(stderr, "bic_init() -- cannot allocate memory\n");
      bic_reset(p);
      return(NULL);
    }

    if ((_ev  = gsl_vector_alloc(dim)) == NULL) {
      fprintf(stderr, "bic_init() -- cannot allocate memory\n");
      bic_reset(p);
      return(NULL);
    }

    if ((_ws = gsl_eigen_symm_alloc(dim)) == NULL) {
      fprintf(stderr, "bic_init() -- cannot allocate memory\n");
      bic_reset(p);
      return(NULL);
    }
    _ctype = BIC_FULL_COV;
  }
  else {
    if ((_dm = (double *)malloc(dim * sizeof(double))) == NULL) {
      fprintf(stderr, "bic_init() -- cannot allocate memory\n");
      bic_reset(p);
      return(NULL);
    }
    if ((_dS = (double *)malloc(dim * sizeof(double))) == NULL) {
      fprintf(stderr, "bic_init() -- cannot allocate memory\n");
      bic_reset(p);
      return(NULL);
    }
  }
#else
  if (fullc) {
    fprintf(stderr, "bic_init() error -- cannot initialize BIC segmentation with full covariance\n");
    fprintf(stderr, "                    build audioseg with GSL support for full covariance\n");
    return(NULL);
  }
  if ((_dm = (double *)malloc(dim * sizeof(double))) == NULL) {
    fprintf(stderr, "bic_init() -- cannot allocate memory\n");
    bic_reset(p);
    return(NULL);
  }
  if ((_dS = (double *)malloc(dim * sizeof(double))) == NULL) {
    fprintf(stderr, "bic_init() -- cannot allocate memory\n");
    bic_reset(p);
    return(NULL);
  }
#endif

  return(p);
}

/* -------------------------------------- */
/* ----- void bic_reset(bicwin_t *) ----- */
/* -------------------------------------- */
/*
 * Free memory.
 */
void bic_reset(bicwin_t *p)
{
  if (p) {
    if (p->buf)
      spf_buf_free(p->buf);
    free(p);
  }

#ifdef HAVE_GSL
  if (_ctype == BIC_FULL_COV) {
    /* free memory allocated to GSL stuff */
    if (_m)
      gsl_vector_free(_m); 
    if (_S)
      gsl_matrix_free(_S);
    if (_ws)
      gsl_eigen_symm_free(_ws);
    if (_ev)
      gsl_vector_free(_ev);
  }
  else {
    if (_dm)
      free(_dm);
    if (_dS)
      free(_dS);
  }
#else
  if (_dm)
    free(_dm);
  if (_dS)
    free(_dS);
#endif
}

/* ------------------------------------------ */
/* ----- double dbic(bicwin_t *, float) ----- */
/* ------------------------------------------ */
/*
 * Compute \Delta\mbox{BIC} for the specified window.
 */
double dbic(bicwin_t *win, float lambda)
{
  double s, s1, s2;
  unsigned long st, n, n1, n2;

  double log_det(spfbuf_t *, unsigned long, unsigned long);

  /* ----- entire window ----- */
  st = win->is - win->offset;
  n = win->ie - win->is;
  s = log_det(win->buf, st, n); 

  /* ----- 1st sub-window ----- */
  st = win->is - win->offset;
  n1 = win->im - win->is;
  s1 = log_det(win->buf, st, n1); 

  /* ----- 2nd sub-window ---- */
  st = win->im - win->offset;
  n2 = win->ie - win->im;
  s2 = log_det(win->buf, st, n2); 

  return(0.5 * (n1 * s1 + n2 * s2 - n * s + lambda * win->cp * log(n)));
}

/* ------------------------------------------------------------------------------------ */
/* ----- int bic_change_search(bicwin_t *, unsigned long, float, unsigned long *) ----- */
/* ------------------------------------------------------------------------------------ */
/* 
 * Search for an hypothesized boundary in the specified window. If a
 * change is found, set the frame index of the first frame after the
 * change and return 1. Return 0 otherwise.
 */
int bic_change_search(bicwin_t *win, unsigned long inc, float lambda, unsigned long *imin) 
{
  double d, dmin, penalty;
  double s, s1, s2;
  unsigned long st, n, n1, n2;

  double log_det(spfbuf_t *, unsigned long, unsigned long);


  /* ----- determinant for the entire window ----- */
  st = win->is - win->offset;
  if ((n = win->ie - win->is) < inc)
    return(0);
  s = log_det(win->buf, st, n); 

  penalty = lambda * win->cp * log(n);

  /* ----- loop on all mid positions ----- */
  win->im = win->is + inc;
  *imin = 0;
  dmin = DBL_MAX;

  while (win->im < win->ie) {

    /* 1st sub-window */
    st = win->is - win->offset;
    n1 = win->im - win->is;
    s1 = log_det(win->buf, st, n1);
    
    /* 2nd sub-window */
    st = win->im - win->offset;
    n2 = win->ie - win->im;
    s2 = log_det(win->buf, st, n2);
    
    d = 0.5 * (n1 * s1 + n2 * s2 - n * s + penalty);

    if (d < dmin) {
      *imin = win->im;
      dmin = d;
    }
    win->im += inc;
  }
 
  /*
    if (dmin != DBL_MAX)
    fprintf(stderr, "search: start=%lu  end=%lu  inc=%lu --> imin=%d (%.4f)\n", 
    win->is, win->ie, inc, dmin > 0 ? -1 : (int)(*imin), dmin);
  */

  if (dmin > 0)
    return(0);

  return(1);
}

/* -------------------------------------------------------------------- */
/* ----- double log_det(spfbuf_t *, unsigned long, unsigned long) ----- */
/* -------------------------------------------------------------------- */
/*
 * Compute log-determinant.
 */
double log_det(spfbuf_t *buf, unsigned long st, unsigned long n)
{
  unsigned long i;
  unsigned short j, k;
  double z = (double)1.0 / (double)n;
  double zz = z * z;
  double *vp, *mp;
  double a, logd;

  spf_t *yt = get_spf_buf_vec(buf, st);
  unsigned short dim = spf_buf_dim(buf);

#ifdef HAVE_GSL

  if (_ctype == BIC_FULL_COV) {
    
    mp = gsl_vector_ptr(_m, 0);

    /* accumulate sufficient statistics */
    gsl_vector_set_zero(_m);
    gsl_matrix_set_zero(_S);

    for (i = 0; i < n; i++) {

      for (j = 0; j < dim; j++) {
	vp = gsl_matrix_ptr(_S, j, j);
	a = (double)*(yt+j);
      
	*(mp+j) += a;
	*vp += a * a;

	for (k = 1; k < dim - j; k++) {
	  vp += dim;
	  *vp += a * (double)*(yt+j+k);
	}
      }
      
      yt += buf->adim;
    }
  
    /* compute variance */
    for (j = 0; j < dim; j++) {
      vp = gsl_matrix_ptr(_S, j, j);
      
      a = *vp * z - *(mp+j) * *(mp+j) * zz;
      *vp = a;

      for (k = 1; k < dim - j; k++) {
	vp += dim;
	*vp = *vp * z - *(mp+j) * *(mp+j+k) * zz;
      }
    }

    /* compute eigen values and determinant */
    gsl_eigen_symm(_S, _ev, _ws);

    logd = 0.0;
    vp = gsl_vector_ptr(_ev, 0);
    for (j = 0; j < dim; j++)
      logd += log(*(vp+j));
  }
  else { /* BIC_DIAG_COV */
    mp = _dm;
    vp = _dS;
    
    for (j = 0; j < dim; j++) {
      *(mp+j) = 0.0;
      *(vp+j) = 0.0;
    }
    
    for (i = 0; i < n; i++) {
      
      for (j = 0; j < dim; j++) {
	a = (double)*(yt+j);
	
	*(mp+j) += a;
	*(vp+j) += a * a;
      }
      
      yt += buf->adim;
    }

    logd = 0.0;
    for (j = 0; j < dim; j++) {
      a = *(mp+j);
      logd += log(*(vp+j) * z - a * a * zz);
    }
  }
#else
  mp = _dm;
  vp = _dS;
    
  for (j = 0; j < dim; j++) {
    *(mp+j) = 0.0;
    *(vp+j) = 0.0;
  }
    
  for (i = 0; i < n; i++) {
    
    for (j = 0; j < dim; j++) {
      a = (double)*(yt+j);
      
      *(mp+j) += a;
      *(vp+j) += a * a;
    }
      
    yt += buf->adim;
  }

  logd = 0.0;
  for (j = 0; j < dim; j++) {
    a = *(mp+j);
    logd += log(*(vp+j) * z - a * a * zz);
  }
#endif

  return(logd);
}

/* -------------------------------------------------------------------------------- */
/* ----- unsigned long bic_win_fill(bicwin_t *, unsigned long, unsigned long) ----- */
/* -------------------------------------------------------------------------------- */
/*
 * bic_win_fill(win, start_index, nframes)
 *
 * fill in the BIC window buffer starting at frame start_index with
 * nframes frames or up to the window max_index or the end of the
 * window input stream, which ever comes first. The buffer is extended
 * as necessary if not enough memory has been allocated.
 *
 * Returns the actual window length (i.e. nframes or less if the
 * max_index limit or if the end of stream was reached before reading
 * nframes) or zero in case of error.
 *
 * NOTE: start and max indexes are *absolute* frame indexes.
 */
unsigned long bic_win_fill(bicwin_t *p, unsigned long start, unsigned long nframes)
{
  unsigned long i, n, nadded;
  unsigned short j, dim = spf_stream_dim(p->f);
  spf_t *yt, *bp;

  /* fprintf(stderr, "fill: [%lu %lu] start=%lu nframes=%lu -> ", p->is, p->ie, start, nframes); */

  /* reset buffer -- if start is between the current window, reuse the
     frames already in the window buffer and read the remaining ones,
     if any. */
  n = 0;
  if (start >=  p->is && start <= p->ie) {
    for (i = start - p->offset; i < p->buf->n; i++) {
      if (i != n) {
	yt = get_spf_buf_vec(p->buf, i);
	bp = get_spf_buf_vec(p->buf, n);
	
	for (j = 0; j < dim; j++) 
	  *(bp+j) = *(yt+j);
      }
      n++;
      
      if (n == nframes || start + n == p->maxi)
	break;
    }
  }
  p->is = start;
  p->offset = start;
  p->ie = start + n;
  p->buf->n = n;

  /* fprintf(stderr, "%lu", n); */

  if (n == nframes) {
    /* fprintf(stderr, " + 0 = %lu   (offset=%lu, maxi=%lu)\n", p->buf->n, p->offset, p->maxi); */
    return(n);
  }
  if (p->maxi > 0 && start + n >= p->maxi) {
    p->ie = p->maxi;
    /* fprintf(stderr, " + 0 = %lu   (offset=%lu, maxi=%lu)\n", p->buf->n, p->offset, p->maxi); */
    return(n);
  }

  /* if we need some more frames, it's time to seek into the stream
     and append the required number of frames */
  if (spf_stream_seek(p->f, start + n, SEEK_SET) != 0)
      /* if cannot seek to append data, then use whatever we have from
	 the previous buffer */
      return(p->buf->n);

  /* now get more frames */
  nadded = 0;
  while ((yt = get_next_spf_frame(p->f)) != NULL) {

    if (spf_buf_append(p->buf, yt, spf_stream_dim(p->f), 500) == NULL) {
      /* fprintf(stderr, "bic_win_fill() -- cannot add frame to buffer\n"); */
      return(0);
    }
    
    nadded++;

    if ((p->maxi > 0 && p->ie + nadded == p->maxi) || nadded == nframes - n)
      break;
  }

  p->ie += nadded;

  /* if we reached the end of the stream, set max index so that we
     won't try to seek past the end of stream next time we go into
     bic_win_fill(). */
  if (! yt)
    p->maxi = p->ie;

  /* fprintf(stderr, " + %lu = %lu   (offset=%lu, maxi=%lu)\n", nadded, p->buf->n, p->offset, p->maxi); */

  return(p->buf->n);
}

#undef _bic_c_
