/******************************************************************************/
/*                                                                            */
/*                                cluster.h                                   */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Guig                                                              Aug 2003 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2003 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 *   $Author: ggravier $
 *   $Date: 2003/08/22 14:15:10 $
 *   $Revision: 1.2 $
 *
 */

/*
 * BIC clustering utilities.
 */

#ifndef _cluster_h_
# define _cluster_h_

# include <audioseg.h>

# define CLUSTER_NAME_STRING "C"

     /* ---------------------------- */
     /* ----- type definitions ----- */
     /* ---------------------------- */

/*
 * Link to segments used by clusters to keep track of their
 * guys. Links are used to represent a collection of segments as a
 * d-list.
 */
typedef struct _seglink_s {
  asseg_t *seg;                 /* segment address of the link element        */ 
  struct _seglink_s *next;      /* next link element                          */ 
} seglink_t;

/*
 * Clusters are organized as a d-list fo fast and easy merge/delete
 * operations. As a consequence, keeping track of the clustering tree
 * structure is impossible whithin the cluster list and a different
 * structure should be used for this purpose if we'd want to keep
 * track of the tree structure (which not something we wanna
 * do). However, the current cluster representation saves a lot of
 * memory by avoiding duplication of the mean and variance accumulator
 * as one would do if he'd had to keep all the tree structure.
 */
typedef struct _cluster_s {
  unsigned short dim;           /* dimension (redundant but usefull)          */
  unsigned long nframes;        /* total number of frames in cluster          */ 
  double *m;                    /* mean accumulator                           */
  double **v;                   /* variance accumulator                       */
  double logdet;                /* log-determinant of the covariance matrix   */
  unsigned long nsegs;          /* number of associated segments              */
  seglink_t *link;              /* link to associated segments                */
  struct _cluster_s *next;      /* next cluster                               */
  struct _cluster_s *prev;      /* next cluster                               */
} cluster_t;


/*
 * Inter-cluster distance, elements of the upper triangular
 * inter-cluster distance matrix. As clusters have no ID number in the
 * d-list (they may have one but that would not give us fast access to
 * the cluster -- an array would be necessary for fast access but
 * then, merging and deleting clusters would become complex
 * operations), each matrix element must keep track of the
 * corresponding clusters address. Therefore, each row of the matrix
 * has an associated cluster address, and each element in a row
 * contains the address of the second cluster as well as the distance
 * and a flag indicating whether this couple of clusters has already
 * been considered or not. The flag is used mostly to find the next
 * best couple of clusters for which dbic is negative.
 */
typedef struct {
  cluster_t *c2;                /* column associated cluster                  */
  char flag;                    /* discarded or not?                          */
  double d;                     /* actual distance                            */
} icdist_t;

typedef struct {
  unsigned long n;              /* actual number of rows                      */
  unsigned long nmax;           /* allocated number of rows                   */
  struct _icdistrow_s {
    cluster_t *c1;              /* row associated cluster                     */
    unsigned long n;            /* row actual size                            */
    icdist_t *elt;              /* row elements                               */
  } *row;
} icdistmat_t;

     /* ------------------------------- */
     /* ----- function prototypes ----- */
     /* ------------------------------- */

/*
 * Initialize workspace for clustering.
 */
int cluster_init(
  unsigned short,               /* dimension                                  */
  int                           /* use full covariance (0 for diag. cov.)     */
);

/*
 * Reset clustering workspace.
 */
void cluster_reset(void);

/*
 * Allocate cluster.
 */
cluster_t *cluster_alloc(
  unsigned short                /* dimension                                  */
);

/*
 * Free cluster.
 */
void cluster_free(
  cluster_t *                   /* cluster address                            */
);

/*
 * Free cluster list
 */
void cluster_list_free(
  cluster_t *                   /* list of clusters                           */  
);

/* 
 * Merge the two clusters into the first one. Return a pointer to
 * the possibly changed list address. 
 */
cluster_t *cluster_merge(
  cluster_t *,                  /* cluster list address                       */
  cluster_t *,                  /* first cluster                              */
  cluster_t *                   /* second cluster                             */
);


/*
 * Compute log-determinant of the covariance matrix for a cluster
 */
double cluster_const_set(
  cluster_t *                   /* cluster list address                       */
);
		      
/*
 * Allocate inter-cluster matrix.
 */
icdistmat_t *distance_matrix_alloc(
  unsigned long                 /* maximum number of clusters                 */
);

/*
 * Free inter-cluster matrix
 */
void distance_matrix_free(
  icdistmat_t *                 /* matrix to free                             */
);

/*
 * Compute inter-cluster distances and return a pointer to the matrix
 * element with the smallest distance.
 */
void distance_matrix_fill(
  icdistmat_t *,                /* matrix to fill                             */
  cluster_t *[2],               /* clusters with the smallest distance        */
  double *,                     /* smallest distance                          */ 
  cluster_t *,                  /* list of clusters                           */
  double(*)(cluster_t *, cluster_t *, float), /* distance computation function  */
  float                         /* parameter of the distance function         */
);

/*
 * Search the unvisited element with the smallest distance in the
 * matrix. Return a pointer to that element.
 */
int distance_matrix_search(
  icdistmat_t *,                /* matrix to search                           */
  cluster_t *[2],               /* clusters with the smallest distance        */
  double *                      /* smallest distance                          */ 
);

/*
 * Return the symmetric KL distance between two clusters
 */
double cluster_kl2_distance(
  cluster_t *,                  /* first cluster                              */
  cluster_t *,                  /* second cluster                             */
  float                         /* unused                                     */
);

/*
 * Return the log-likelihood ratio distance between two clusters
 */
double cluster_like_distance(
  cluster_t *,                  /* first cluster                              */
  cluster_t *,                  /* second cluster                             */
  float                         /* unsued                                     */
);

/*
 * Return the BIC distance between two clusters
 */
double cluster_bic_distance(
  cluster_t *,                  /* first cluster                              */
  cluster_t *,                  /* second cluster                             */
  float                         /* complexity penalty weight                  */
);


/*
 * Return the delta BIC between two clusterings
 */
double cluster_delta_bic(
  cluster_t *,                  /* first cluster                              */
  cluster_t *,                  /* second cluster                             */
  float,                        /* lambda                                     */
  unsigned long                 /* total number of frames                     */
);

# ifdef _cluster_c_

#  define CLUSTER_DIAG_COV 0
#  define CLUSTER_FULL_COV 1

static int _ctype = CLUSTER_DIAG_COV;

#  ifdef HAVE_GSL

#   include <gsl/gsl_blas.h>
#   include <gsl/gsl_eigen.h>

static gsl_matrix *_S = NULL;
static gsl_eigen_symm_workspace *_ws = NULL;
static gsl_vector *_ev = NULL;

#  endif

# endif /* _cluster_c_ */

#endif /* _cluster_h_ */
