/******************************************************************************/
/*                                                                            */
/*                                  gmm.c                                     */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Guig                                                             Aug, 2002 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2003 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 * $Author: ggravier $
 * $Date: 2003/11/20 15:50:58 $
 * $Revision: 1.5 $
 *
 */

/*
 * Gaussian Mixture Modeling Library.
 */

#define _gmm_c_

#include <audioseg.h>

/* ------------------------------------------------------------ */
/* ----- gmm_t *gmm_alloc(unsigned short, unsigned short) ----- */
/* ------------------------------------------------------------ */
/*
 * Allocate memory for a model. Return a pointer to the model or NULL
 * in case of error.
 */
gmm_t *gmm_alloc(unsigned short n, unsigned short dim)
{
  gmm_t *p;
  unsigned short i;
  
  if (n == 0 || dim == 0)
    return(NULL);
  
  if ((p = (gmm_t *)malloc(sizeof(gmm_t))) == NULL)
    return(NULL);
  
  p->n = n;
  p->dim = dim;
  p->w = NULL;
  p->m = NULL;
  p->v = NULL;
  p->logc = NULL;
  p->expz = NULL;

  if ((p->w = (float *)malloc(n * SIZEOF_FLOAT))==NULL) {
    gmm_free(p);
    return(NULL);
  }

  if ((p->m = (float **)malloc(n * sizeof(float *))) == NULL) {
    gmm_free(p);
    return(NULL);
  }
  for (i = 0; i < n; i++)
    *(p->m+i) = NULL;

  if ((p->v = (float **)malloc(n * sizeof(float *))) == NULL) {
    gmm_free(p);
    return(NULL);
  }
  for (i = 0; i < n; i++)
    *(p->v+i) = NULL;

  for (i = 0; i < n; i++) {
    if ((*(p->m+i) = (float *)malloc(dim * SIZEOF_FLOAT)) == NULL) {
      gmm_free(p);
      return(NULL);
    }
    
    if ((*(p->v+i) = (float *)malloc(dim * SIZEOF_FLOAT)) == NULL) {
      gmm_free(p);
      return(NULL);
    }
  }
  
  
  if ((p->logc = (double *)malloc(n * SIZEOF_DOUBLE)) == NULL) {
    gmm_free(p);
    return(NULL);
  }
  if ((p->expz = (double *)malloc(n * SIZEOF_DOUBLE)) == NULL) {
    gmm_free(p);
    return(NULL);
  }

  return(p);
}

/* ----------------------------------- */
/* ----- void gmm_free(sgmm_t *) ----- */
/* ----------------------------------- */
/*
 * Free memory allocated to a GMM.
 */
void gmm_free(gmm_t *p)
{
  unsigned short i;
  
  if (p) {

    if (p->w)
      free(p->w);

    if (p->m)
      for (i = 0; i < p->n; i++) 
	if (*(p->m+i))
	  free(*(p->m+i));

    if (p->v)
      for (i = 0; i < p->n; i++) 
	if (*(p->v+i))
	  free(*(p->v+i));
    
    if (p->logc)
      free(p->logc);

    if (p->expz)
      free(p->expz);

    free(p);
  }
}

/* ---------------------------------------------------------- */
/* ----- void swap_bytes(void *, unsigned long, size_t) ----- */
/* ---------------------------------------------------------- */
/*
 * Swap bytes
 */
void swap_bytes(void *p, unsigned long n, size_t m)
{
  char c, *cp = (char *)p;
  unsigned long i;
  size_t j;
  
  for (i = 0; i < n; i++) {

    for (j = 0; j < m / 2; j++) {
      c = cp[j];
      cp[j] = cp[m-j-1];
      cp[m-j-1] = c;
    }

    cp += m;
  }
}

/* ----------------------------------------- */
/* ----- gmm_t *gmm_read(const char *) ----- */
/* ----------------------------------------- */
/*
 * Read GMM from file.
 */
gmm_t *gmm_read(const char *fn)
{
  FILE *f;
  gmm_t *p;
  unsigned short i, n, dim;

  if (fn && strcmp(fn, "-") != 0) {
    if ((f = fopen(fn, "rb")) == NULL) {
      fprintf(stderr,"gmm_read() -- cannot open file %s\n", fn);
      return(NULL);
    }
  }
  else
    f = stdin;

  if (fread(&n, SIZEOF_SHORT, 1, f) != 1 || fread(&dim, SIZEOF_SHORT, 1 , f) != 1) {
    fprintf(stderr, "gmm_read() -- cannot read header from file %s\n", fn ? fn : "stdin");
    if (f != stdin) fclose(f);
    return(NULL);
  }
#ifdef WORDS_BIGENDIAN
  swap_bytes(&n, 1, SIZEOF_SHORT);
  swap_bytes(&dim, 1, SIZEOF_SHORT);
#endif /* WORDS_BIGENDIAN */

  if ((p = gmm_alloc(n, dim)) == NULL) {
    fprintf(stderr,"gmm_read() -- cannot allocate memory\n");
    if (f != stdin) fclose(f);
    return(NULL);
  }

  if (fread(p->w, SIZEOF_FLOAT, n, f) != n) {
    fprintf(stderr, "gmm_read() -- cannot read weights from file %s\n", fn ? fn : "stdin");
    gmm_free(p); if (f != stdin) fclose(f);
    return(NULL);
  }
#ifdef WORDS_BIGENDIAN
  swap_bytes(p->w, n, SIZEOF_FLOAT);
#endif /* WORDS_BIGENDIAN */

  for (i = 0; i < n; i++) {

    if (fread(*(p->m+i), SIZEOF_FLOAT, dim, f) != dim) {
      fprintf(stderr,"gmm_read() -- cannot read mean vector for component %d\n", i);
      gmm_free(p); if (f != stdin) fclose(f);
      return(NULL);
    }

    if (fread(*(p->v+i), SIZEOF_FLOAT, dim, f) != dim) {
      fprintf(stderr,"gmm_read() -- cannot read variance vector for component %d\n", i);
      gmm_free(p); if (f != stdin) fclose(f);
      return(NULL);
    }
    
#ifdef WORDS_BIGENDIAN
    swap_bytes(*(p->m+i), dim, SIZEOF_FLOAT);
    swap_bytes(*(p->v+i), dim, SIZEOF_FLOAT);
#endif /* WORDS_BIGENDIAN */
  }

  if (f != stdin) 
    fclose(f);

  gmm_const_set(p);

  return(p);
}

/* ------------------------------------------------ */
/* ----- int gmm_write(const char *, gmm_t *) ----- */
/* ------------------------------------------------ */
/*
 * Write model to the file. Return 0 if ok.
 */
int gmm_write(const char *fn, gmm_t *p)
{
  FILE *f;
  unsigned short i, n, dim;

  if (fn && strcmp(fn, "-") != 0) {
    if ((f = fopen(fn, "wb")) == NULL) {
      fprintf(stderr, "gmm_write() -- cannot open file %s\n", fn);
      return(1);
    }
  }
  else
    f = stdout;

  n = p->n;
  dim = p->dim;

#ifdef WORDS_BIGENDIAN
  swap_bytes(&n, 1, SIZEOF_SHORT);
  swap_bytes(&dim, 1, SIZEOF_SHORT);
#endif /* WORDS_BIGENDIAN */

  if (fwrite(&n, SIZEOF_SHORT, 1, f) != 1 || fwrite(&dim, SIZEOF_SHORT, 1, f) != 1) {
    fprintf(stderr, "gmm_write() -- cannot write header to file %s\n", fn ? fn : "stdout");
    if (f != stdout) fclose(f);
    return(1);
  }

#ifdef WORDS_BIGENDIAN
  n = p->n;
  dim = p->dim;
#endif /* WORDS_BIGENDIAN */

#ifdef WORDS_BIGENDIAN
  swap_bytes(p->w, n, SIZEOF_FLOAT);
#endif /* WORDS_BIGENDIAN */

  if (fwrite(p->w, SIZEOF_FLOAT, n, f) != n) {
    fprintf(stderr, "gmm_write() -- cannot write weights to file %s\n", fn ? fn : "stdout");
    if (f != stdout) fclose(f);
    return(1);
  }

#ifdef WORDS_BIGENDIAN
  swap_bytes(p->w, n, SIZEOF_FLOAT);
#endif /* WORDS_BIGENDIAN */

  for (i = 0; i < n; i++) {

#ifdef WORDS_BIGENDIAN
    swap_bytes(*(p->m+i), dim, SIZEOF_FLOAT);
    swap_bytes(*(p->v+i), dim, SIZEOF_FLOAT);
#endif /* WORDS_BIGENDIAN */

    if (fwrite(*(p->m+i), SIZEOF_FLOAT, dim, f) != dim) {
      fprintf(stderr, "gmm_write() -- cannot write mean for component %d\n", i);
      if (f != stdout) fclose(f);
      return(1);
    }

    if (fwrite(*(p->v+i), SIZEOF_FLOAT, dim, f) != dim) {
      fprintf(stderr, "gmm_write() -- cannot write variance for component %d\n", i);
      if (f != stdout) fclose(f);
      return(1);
    }

#ifdef WORDS_BIGENDIAN
    swap_bytes(*(p->m+i), dim, SIZEOF_FLOAT);
    swap_bytes(*(p->v+i), dim, SIZEOF_FLOAT);
#endif /* WORDS_BIGENDIAN */
  }

  if (f != stdout)
    fclose(f);

  return(0);
}

/* ------------------------------------------- */
/* ----- void gmm_print(FILE *, gmm_t *) ----- */
/* ------------------------------------------- */
/*
 * Print a GMM to file in ASCII format.
 */
void gmm_print(FILE *f, gmm_t *gmm)
{
  unsigned short i, j;
  float *p;
  unsigned short ncomp, dim;

  ncomp = get_gmm_num_comp(gmm);
  dim = get_gmm_dim(gmm);

  fprintf(f, "%u %u\n", ncomp, dim);

  for (i = 0; i < ncomp; i++) {

    fprintf(f, "%u %f\n", i, get_gmm_comp_weight(gmm, i));

    p = get_gmm_comp_mean(gmm, i);
    for (j = 0; j < dim; j++)
      fprintf(f, j ? " %-10.5f": "%-10.5f", *(p+j));
    fprintf(f, "\n");

    p = get_gmm_comp_ivar(gmm, i);
    for (j = 0; j < dim; j++)
      fprintf(f, j ? " %-10.5f": "%-10.5f", 1.0 / *(p+j));
    fprintf(f, "\n");
  }
}

/* --------------------------------------- */
/* ----- void gmm_const_set(gmm_t *) ----- */
/* --------------------------------------- */
/*
 * Pre-compute constant terms in likelihodd.
 *
 * For a gaussian of covariance vector v \in \mathbb{R}^{d} and weight w,
 * the constants are given by:
 *
 *   logc = -0.5 * d * \ln(2 \pi) - 0.5 * \ln(\prod_{j = 1}^{d} v_j) + \ln (w)
 *    = \frac{w}{\sqrt{\prod_{j = 1}^{d} v_j}}
 *
 * NOTE: p->v contain the *inverse* covariances!
 */
void gmm_const_set(gmm_t *gmm)
{
  double sum, prod;
  unsigned short i, j, dim = get_gmm_dim(gmm);
  float *vp, w;

  for (i = 0; i < get_gmm_num_comp(gmm); i++) {
    w = get_gmm_comp_weight(gmm, i);

    
    if (w > GMM_MIN_WEIGHT) {
      sum = 0.0;
      prod = 1.0;

      vp = get_gmm_comp_ivar(gmm, i);
      
      for (j = 0; j < dim; j++) {
	prod *= (double)*(vp+j);
	sum += log((double)*(vp+j));
      }
      
      set_gmm_comp_logc(gmm, i, -0.5 * dim * LOG_2_PI + log(w) + 0.5 * sum);
      set_gmm_comp_expz(gmm, i, w * sqrt(prod));
    }
    else {
      set_gmm_comp_logc(gmm, i, 0.0);
      set_gmm_comp_expz(gmm, i, 0.0);
    }
  }
}

/* ------------------------------------------------------- */
/* ----- double gmm_frame_log_like(spf_t *, gmm_t *) ----- */
/* ------------------------------------------------------- */
/*
 * Computes the log-likelihood of a frame for a gaussian mixture model
 */
double gmm_frame_log_like(spf_t *yt, gmm_t *gmm)
{
  unsigned short i, j, ibest;
  static double s[GMM_NUM_COMP_MAX];
  double a, b, sum, sum1, sum2;
  float *m, *v, *wp = get_gmm_weights(gmm);

  unsigned short dim = get_gmm_dim(gmm);

  /* compute sums inside exponentials for each Gaussian */
  ibest = 0;

  for (i = 0; i < get_gmm_num_comp(gmm);  i++) {

    if (*(wp+i) > GMM_MIN_WEIGHT) {
      sum = 0.0;
      m = get_gmm_comp_mean(gmm, i);
      v = get_gmm_comp_ivar(gmm, i);
    
      for (j = 0; j < dim; j++) {
	a = (double)*(yt+j) - (double)*(m+j);
	sum += a * a * (double)*(v+j);
      }

      *(s+i) = sum;

      if (sum < *(s+ibest))
	ibest = i;
    }
    else
      *(s+i) = 0.0;
  }

  /* compute log-likelihood */
  a = 1.0 / get_gmm_comp_expz(gmm, ibest);
  b = *(s+ibest);
  
  sum1 = get_gmm_comp_logc(gmm, ibest) - 0.5 * b;
  sum2 = 1.0;

  for (i = 0; i < get_gmm_num_comp(gmm); i++)
    if (i != ibest && *(wp+i) > GMM_MIN_WEIGHT)
      sum2 += a * get_gmm_comp_expz(gmm, i) * exp(-0.5 * (*(s+i) - b));

  return(sum1 + log(sum2));
}

/* ----------------------------------------------------------------------------------------- */
/* ----- double gmm_nbest_frame_log_like(spf_t *, gmm_t *, int *, unsigned short, int) ----- */
/* ----------------------------------------------------------------------------------------- */
/*
 * Computes the log-likelihood of a frame for a gaussian mixture model
 * using the n-best components.  
 */
double gmm_nbest_frame_log_like(spf_t *yt, gmm_t *gmm, int *idx, unsigned short nbest, int set)
{
  unsigned short ibest, i, j, k, dim = get_gmm_dim(gmm);
  double a, b, sum, sum1, sum2;
  float *m, *v, *wp = get_gmm_weights(gmm);
  struct {
    unsigned short index;
    double sum;
    double logp;
  } s[GMM_NUM_COMP_MAX];

  int _lcmp(const void *, const void *);

  if (set) {

    for (i = 0; i < get_gmm_num_comp(gmm);  i++) {
    
      if (*(wp+i) > GMM_MIN_WEIGHT) {
	sum = 0.0;
	m = get_gmm_comp_mean(gmm, i);
	v = get_gmm_comp_ivar(gmm, i);
    
	for (j = 0; j < dim; j++) {
	  a = (double)*(yt+j) - (double)*(m+j);
	  sum += a * a * (double)*(v+j);
	}

	(s+i)->sum = sum;
	(s+i)->logp = get_gmm_comp_logc(gmm, i) - 0.5 * sum;
      }
      else {
	(s+i)->sum = DBL_MAX;
	(s+i)->logp = -DBL_MAX;
      }

      (s+i)->index = i;
    }

    /* sort logprobs array */
    qsort(s, get_gmm_num_comp(gmm), sizeof(s[0]), _lcmp);

    /* set idx array */
    for (k = 0; k < nbest;  k++)
      *(idx+k) = (s+k)->index;

  }
  else
    for (k = 0; k < nbest;  k++) {
      i = *(idx+k);

      if (*(wp+i) > GMM_MIN_WEIGHT) {
	sum = 0.0;
	m = get_gmm_comp_mean(gmm, i);
	v = get_gmm_comp_ivar(gmm, i);
	
	for (j = 0; j < dim; j++) {
	  a = (double)*(yt+j) - (double)*(m+j);
	  sum += a * a * (double)*(v+j);
	}

	(s+k)->sum = sum;
      }
      else
	(s+k)->sum = DBL_MAX;
      
      (s+k)->index = i;
    }

  /* compute log-likelihood */
  ibest = s->index;
  b = s->sum;
  a = 1.0 / get_gmm_comp_expz(gmm, ibest);

  sum1 = get_gmm_comp_logc(gmm, ibest) - 0.5 * b;
  sum2 = 1.0;

  for (k = 1; k < nbest; k++) {
    i = *(idx+k);

    if (*(wp+i) > GMM_MIN_WEIGHT)
      sum2 += a * get_gmm_comp_expz(gmm, i) * exp(-0.5 * ((s+k)->sum - b));
  }

  return(sum1 + log(sum2));
}

int _lcmp(const void *a, const void *b)
{
  struct score_s {
    unsigned short index;
    double sum;
    double logp;
  } *x, *y;
  int d = 0;

  x = (struct score_s *)a;
  y = (struct score_s *)b;

  if (x->logp > y->logp)
    d = -1;
  else if (x->logp < y->logp)
    d = 1;
  
  return(d);
}

/* -------------------------------------------------------------------------------- */
/* ----- double gmm_segment_log_like(spfstream_t *, unsigned long *, gmm_t *) ----- */
/* -------------------------------------------------------------------------------- */
/*
 * Return the sum of the log-likelihoods over all frames.
 */
double gmm_segment_log_like(spfstream_t *f, unsigned long *nframes, gmm_t *gmm)
{
  double sum;
  unsigned long t;
  spf_t *yt;
  
  t = 0;
  sum = 0.0;

  while ((yt = get_next_spf_frame(f)) != NULL) {
    
    sum += gmm_frame_log_like(yt, gmm);
    t++;
    if (t == *nframes)
      break;
  }

  *nframes = t;

  if (t == 0)
    return(-DBL_MAX);

  return(sum);
}

#undef _gmm_c_
