/******************************************************************************/
/*                                                                            */
/*                                  script.c                                  */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Guig                                                       Fri Nov 15 2002 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2002 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 *   $Author: ggravier $
 *   $Date: 2003/11/20 15:50:58 $
 *   $Revision: 1.5 $
 *
 */

/*
 * Script related functions -- see script.h for comments.
 */

#define _script_c_

#include <script.h>

# define SCRIPT_MAX_LINE_LEN 1024
# define SCRIPT_COMMENT_CHAR '#'

/* ---------------------------------------------------- */
/* ----- script_t *load_script(const char *, int) ----- */
/* ---------------------------------------------------- */
/*
 * Load script from file.
 */
script_t *load_script(const char *fn, int loadseg)
{
  FILE *f;
  char line[SCRIPT_MAX_LINE_LEN];
  char *p;
  int lino = 0;
  script_t *script;
 
   if ((script = (script_t *)malloc(sizeof(script_t))) == NULL) {
    fprintf(stderr, "load_script() -- cannot allocate %d bytes\n", sizeof(script_t));
    return(NULL);
  }
  script->nfiles = 0;
  script->datfn = script->segfn = (char **)NULL;
  script->seg = (asseg_t **)NULL;

  if ((f = fopen(fn, "r")) == NULL) {
    fprintf(stderr, "load_script() -- cannot open script file %s\n", fn);
    free_script(script);
    return(NULL);
  }

  while (fgets(line, SCRIPT_MAX_LINE_LEN, f) != NULL) {
    lino++;

    /* remove comments - i.e. search the # in the line */
    if ((p = (char *)strchr(line, SCRIPT_COMMENT_CHAR)) != NULL)
      *p = 0x00; /* set end of string there */
    
    if (! *line) /* skip empty lines */
      continue;
 
    /* grow the script list */
    if ((script->datfn = (char **)realloc(script->datfn, sizeof(char *) * (script->nfiles + 1))) == NULL) {
      fprintf(stderr, "load_script() -- cannot allocate %ld bytes\n", sizeof(char *) * (script->nfiles + 1));
      free_script(script);
      return(NULL);
    }
    script->datfn[script->nfiles] = NULL;

    if ((script->segfn = (char **)realloc(script->segfn, sizeof(char *) * (script->nfiles + 1))) == NULL) {
      fprintf(stderr, "load_script() -- cannot allocate %ld bytes\n", sizeof(char *) * (script->nfiles + 1));
      free_script(script);
      return(NULL);
    }
    script->segfn[script->nfiles] = NULL;

    if (loadseg) {
      if ((script->seg = (asseg_t **)realloc(script->seg, sizeof(asseg_t *) * (script->nfiles + 1))) == NULL) {
	fprintf(stderr, "load_script() -- cannot allocate %ld bytes\n", sizeof(asseg_t *) * (script->nfiles + 1));
	free_script(script);
	return(NULL);
      }
      script->seg[script->nfiles] = NULL;
    }
    script->nfiles += 1;

    /* fill in the new slot */
    if ((p = strtok(line, " \t\n")) == NULL) {
      fprintf(stderr, "load_script() -- cannot read data filename at line %d in script %s\n", lino, fn);
      free_script(script);
      return(NULL);
    }
    if ((script->datfn[script->nfiles - 1] = strdup(p)) == NULL) {
      fprintf(stderr, "load_script() -- cannot allocate %d bytes\n", strlen(p) * SIZEOF_CHAR);
      free_script(script);
      return(NULL);
    }

    if ((p = strtok(NULL, " \t\n")) != NULL)
      if ((script->segfn[script->nfiles - 1] = strdup(p)) == NULL) {
	fprintf(stderr, "load_script() -- cannot allocate %d bytes\n", strlen(p) * SIZEOF_CHAR);
	free_script(script);
	return(NULL);
      }

    if (loadseg) {
      if (! p) {
	fprintf(stderr, "load_script() -- segment filename not specified at line %d in script %s\n", lino, fn);
	free_script(script);
	return(NULL);
      }
      if ((script->seg[script->nfiles - 1] = seg_read(p)) == NULL) {
	fprintf(stderr, "load_script() -- cannot read segmentation from file %s\n", p);
	free_script(script);
	return(NULL);
      }
    }
  }

  fclose(f);

  if (script->nfiles == 0) {
    fprintf(stderr, "load_script() -- empty training script file %s\n", fn);
    free_script(script);
    return(NULL);
  }

  return(script);
}

/* ----------------------------------------------- */
/* ----- script_t *make_script(int, char **) ----- */
/* ----------------------------------------------- */
/*
 * Make script from array of data filename.
 */
script_t *make_script(int nfiles, char **tab)
{
  script_t *p;
  int i;

  if ((p = (script_t *)malloc(sizeof(script_t))) == NULL) {
     fprintf(stderr, "make_script() -- cannot allocate memory\n");
     return(NULL);
  }
  p->nfiles = nfiles;
  p->datfn = p->segfn = (char **)NULL;
  p->seg = (asseg_t **)NULL;
  
  if ((p->datfn = (char **)malloc(nfiles * sizeof(char *))) == NULL) {
    fprintf(stderr, "make_script() -- cannot allocate %d bytes\n", nfiles * sizeof(char *));
    free_script(p);
    return(NULL);
  }
  for (i = 0; i < nfiles; i++)
    p->datfn[i] = NULL;

  if ((p->segfn = (char **)malloc(nfiles * sizeof(char *))) == NULL) {
    fprintf(stderr, "make_script() -- cannot allocate %d bytes\n", nfiles * sizeof(char *));
    free_script(p);
    return(NULL);
  }
  for (i = 0; i < nfiles; i++)
    p->segfn[i] = NULL;

  for (i = 0; i < nfiles; i++)
    if ((p->datfn[i] = strdup(tab[i])) == NULL) {
      fprintf(stderr, "make_script() -- cannot allocate %d bytes\n", strlen(tab[i]) * SIZEOF_CHAR);
      free_script(p);
      return(NULL);
    }

  return(p);
}

/* ---------------------------------------- */
/* ----- void free_script(script_t *) ----- */
/* ---------------------------------------- */
/*
 * Free memory allocated to a script.
 */
void free_script(script_t *p)
{
  int i;

  if (p) {
    if (p->datfn) {
      for (i = 0; i < p->nfiles; i++)
	if (p->datfn[i])
	  free(p->datfn[i]);
      free(p->datfn);
    }
    if (p->segfn) {
      for (i = 0; i < p->nfiles; i++)
	if (p->segfn[i])
	  free(p->segfn[i]);
      free(p->segfn);
    }
    if (p->seg) {
      for (i = 0; i < p->nfiles; i++)
	if (p->seg[i])
	  seg_list_free(p->seg[i]);
      free(p->seg);
    }
    free(p);
  }
}

/* ------------------------------------------------- */
/* ----- asseg_t **load_script_seg(script_t *) ----- */
/* ------------------------------------------------- */
/*
 * Load segmentations for the script.
 */
asseg_t **load_script_seg(script_t *p)
{
  unsigned long i;

  if ((p->seg = (asseg_t **)malloc(p->nfiles * sizeof(asseg_t *))) == NULL) {
    fprintf(stderr, "load_script_seg() -- cannot allocate %ld bytes\n", p->nfiles * sizeof(asseg_t *));
    return(NULL);
  }

  for (i = 0; i < p->nfiles; i++)
    p->seg[i] = (asseg_t *)NULL;

  for (i = 0; i < p->nfiles; i++) {
    if (p->segfn[i] == NULL) {
      fprintf(stderr, "load_script_seg() -- no segmentation file name for file %s\n", p->datfn[i]);
      return(NULL);
    }
    if ((p->seg[i] = seg_read(p->segfn[i])) == NULL) {
      fprintf(stderr, "load_script_seg() -- cannot load segmentation from file %s\n", p->segfn[i]);
      return(NULL);
    }
  }

  return(p->seg);
}

#undef _script_c_

