/******************************************************************************/
/*                                                                            */
/*                                sgmcopy.c                                   */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Guig                                                             Aug, 2002 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2002 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 * $Author: ggravier $
 * $Date: 2003/11/20 15:50:58 $
 * $Revision: 1.6 $
 *
 */

/*
 * Copy a model.
 *
 * This is a rather trivial function for the moment, mainly intended
 * to print a model in a viewable format or to import models into
 * audioseg. 
 *
 * The ASCII format model is the following:
 *
 * ncomp dim
 * 0 w[0]
 * m[0][0]           ...     m[0][dim-1]
 * v[0][0]           ...     v[0][dim-1]
 * ...
 * ncomp-1 w[ncomp-1]
 * m[ncomp-1][0]     ...     m[ncomp-1][dim-1]
 * v[ncomp-1][0]     ...     v[ncomp-1][dim-1]
 *
 * The constants c1 and c2 are only used for output and must not be in
 * the input file when importing models. To import models, they must
 * be converted into the ASCII format and copied using sgmcopy with
 * the --import option.
 *
 * In the future, operations on models should be added to this
 * function.
 */

#define _sgmcopy_c_

#include <audioseg.h>
#include <getopt.h>

static char *cvsid = "$Id: sgmcopy.c,v 1.6 2003/11/20 15:50:58 ggravier Exp $";

/* ----------------------------------------------- */
/* ----- global variables set by read_args() ----- */
/* ----------------------------------------------- */
int import = 0;                /* import from ASCII format             */
int export = 0;                /* export to ASCII format               */
int combine = 0;               /* combine models in a single one       */
char *ofn = NULL;              /* output filename                      */
int trace = 0;                 /* trace level                          */
int help = 0;                  /* print usage                          */
int version = 0;               /* version flag                         */

/* ---------------------------- */
/* ----- void usage(void) ----- */
/* ---------------------------- */
void usage(void)
{
  fprintf(stdout, "Usage:\n"); 
  fprintf(stdout, "  sgmcopy [options] model ...\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Synopsis:\n");
  fprintf(stdout, "  Copy a model with possible format convertion.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Options:\n");
  fprintf(stdout, "  -i, --import                 import model(s) from ASCII file\n");
  fprintf(stdout, "  -x, --export                 export model(s) to ASCII file\n");
  fprintf(stdout, "  -c, --combine                combine input models\n");
  fprintf(stdout, "  -o, --output=fn              output filename (stdout)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -v, --verbose[=n]            verbose mode\n");
  fprintf(stdout, "  -V, --version                print version number\n");
  fprintf(stdout, "  -h, --help                   this help message\n");
  fprintf(stdout, "\n");
  fflush(stdout);
}

/* ---------------------------------- */
/* ----- int main(int, char **) ----- */
/* ---------------------------------- */
int main(int argc, char** argv)
{
  gmm_t **gmm;
  char *ifn;
  int i, nmodels = 0;
  int read_args(int, char **);
  int gmm_ascii_write(const char *, gmm_t *);
  gmm_t *gmm_ascii_read(const char *);
  gmm_t *gmm_combine(gmm_t **, int);

  /* ----- process command line ----- */
  if (read_args(argc, argv))
    return(1);

  if (help) {
    usage();
    return(0);
  }

  if (version) {
    fprintf(stdout, "audioseg %s -- %s\n\n", VERSION, cvsid);
    return(0);
  }

  if (optind > argc) {
    fprintf(stderr, "sgmcopy error -- no input model specified (use --help to get usage)\n");
    return(1);
  }

  /* ----- loop on all input files ----- */
  if (combine)
    nmodels = argc - optind;
  else
    nmodels = 1;

  if ((gmm = (gmm_t **)malloc(nmodels * sizeof(gmm_t *))) == NULL) {
    fprintf(stderr, "sgmcopy error -- cannot allocate memory\n");
    return(-1);
  }

  i = 0;
  
  while (optind < argc) {

    ifn = argv[optind++];

    if (trace) {
      fprintf(stdout, "file=%s\n", ifn);
      fflush(stdout);
    }
    
    /* read model */
    if (import) 
      gmm[i] = gmm_ascii_read(ifn);
    else
      gmm[i] = gmm_read(ifn);

    if (gmm[i] == NULL) {
      fprintf(stderr, "sgmcopy error -- cannot load model from file %s\n", ifn);
      while (i) gmm_free(gmm[--i]); free(gmm);
      return(1);
    }
    
    if (combine)
      i++;
    else {
      if (export) {
	if (gmm_ascii_write(ofn, gmm[0]) != 0) {
	  fprintf(stderr, "sgmcopy error -- cannot export model to file %s\n", ofn ? ofn : "stdout");
	  gmm_free(gmm[0]); free(gmm);
	  return(1);
	}
      }
      else if (gmm_write(ofn, gmm[0]) != 0) {
	fprintf(stderr, "sgmcopy error -- cannot save model to file %s\n", ofn ? ofn : "stdout");
	gmm_free(gmm[0]); free(gmm);
	return(1);
      }

      gmm_free(gmm[0]);
    }
  }

  if (combine) {
    gmm_t *ogmm;

    if ((ogmm = gmm_combine(gmm, nmodels)) == NULL) {
      fprintf(stderr, "sgmcopy error -- cannot combine models\n");
      while (i) gmm_free(gmm[--i]); free(gmm);
      return(1);      
    }
    if (export) {
      if (gmm_ascii_write(ofn, ogmm) != 0) {
	fprintf(stderr, "sgmcopy error -- cannot export model to file %s\n", ofn ? ofn : "stdout");
	gmm_free(gmm[0]); free(gmm); gmm_free(ogmm);
	return(1);
      }
    }
    else if (gmm_write(ofn, ogmm) != 0) {
      fprintf(stderr, "sgmcopy error -- cannot save model to file %s\n", ofn ? ofn : "stdout");
      gmm_free(gmm[0]); free(gmm); gmm_free(ogmm);
      return(1);
    }
    gmm_free(ogmm);
  }

  free(gmm);
  return(0);
}

/* --------------------------------------------- */
/* ----- gmm_t *gmm_combine(gmm_t **, int) ----- */
/* --------------------------------------------- */
/*
 * Combine input models with equal priors
 */
gmm_t *gmm_combine(gmm_t **list, int ngmm)
{
  gmm_t *p;
  unsigned short dim, ncomp = 0;
  int i, ii, j;
  double z;

  for (i = 0; i < ngmm; i++)
    ncomp += get_gmm_num_comp(list[i]);
  dim = get_gmm_dim(list[0]);
  z = 1.0 / (double)ngmm;

  p = gmm_alloc(ncomp, dim);

  if (p)
    for (i = 0, ii = 0; i < ngmm; i++)
      for (j = 0; j < get_gmm_num_comp(list[i]); j++, ii++) {
	p->w[ii] = list[i]->w[j] * z;
	memcpy(p->m[ii], list[i]->m[j], dim * SIZEOF_FLOAT);
	memcpy(p->v[ii], list[i]->v[j], dim * SIZEOF_FLOAT);
      }
  
  return(p);
}

/* ------------------------------------------------------ */
/* ----- int gmm_ascii_write(const char *, gmm_t *) ----- */
/* ------------------------------------------------------ */
/*
 * Write model in ASCII format.
 */
int gmm_ascii_write(const char *fn, gmm_t *p)
{
  FILE *f;

  if (fn && strcmp(fn, "-") != 0) {
    if ((f = fopen(fn, "wt")) == NULL)
      return(1);
  }
  else
    f = stdout;
  
  gmm_print(f, p);

  if (f != stdout)
    fclose(f);

  return(0);
}

/* ------------------------------------------------ */
/* ----- gmm_t *gmm_ascii_read(const char *) ----- */
/* ------------------------------------------------ */
/*
 * Read model from ASCII format.
 *
 * WARNING: this is a rather *unsafe* function !!!!!
 */
gmm_t *gmm_ascii_read(const char *fn)
{
  FILE *f;
  gmm_t *p;
  int ncomp, dim, i, j;
  float x;

  if (fn && strcmp(fn, "-") != 0) {
    if ((f = fopen(fn, "rt")) == NULL)
      return(NULL);
  }
  else
    f = stdin;

  if (fscanf(f, "%d %d", &ncomp, &dim) != 2) {
    fprintf(stderr, "sgmcopy error -- cannot read ASCII header\n");
    return(NULL);
  }

  if ((p = gmm_alloc(ncomp, dim)) == NULL) {
    fprintf(stderr, "sgmcopy error -- cannot allocate memory\n");
    return(NULL);
  }

  for (i = 0; i < ncomp; i++) {
    if (fscanf(f, "%d %f", &j, &x) != 2) {
      fprintf(stderr, "sgmcopy error -- cannot read component %d ASCII header\n", i);
      gmm_free(p);
      return(NULL);
    }
    p->w[i] = x;

    for (j = 0; j < dim; j++) {
      if (fscanf(f, "%f", &x) != 1) {
	fprintf(stderr, "sgmcopy error -- cannot read component %d mean[%d]\n", i, j);
	gmm_free(p);
	return(NULL);
      }
      p->m[i][j] = x;
    }

    for (j = 0; j < dim; j++) {
      if (fscanf(f, "%f", &x) != 1) {
	fprintf(stderr, "sgmcopy error -- cannot read component %d mean[%d]\n", i, j);
	gmm_free(p);
	return(NULL);
      }
      p->v[i][j] = 1.0 / x;
    }
  }

  gmm_const_set(p);

  if (f != stdin)
    fclose(f);

  return(p);
}

/* --------------------------------------- */
/* ----- int read_args(int, char **) ----- */
/* --------------------------------------- */
/*
 *  -v, --verbose         verbose mode
 *  -V, --version         print version number
 *  -h, --help            help
 */
int read_args(int argc, char **argv)
{
  int c, i;
  struct option longopts[] = {
    {"import", no_argument, NULL, 'i'},
    {"export", no_argument, NULL, 'x'},
    {"output", required_argument, NULL, 'o'},
    {"verbose", no_argument, NULL, 'v'},
    {"version", no_argument, NULL, 'V'},
    {"help", no_argument, NULL, 'h'},
    {0, 0, 0, 0}
  };

  opterr = 0;
  optopt = '!';

  while ((c = getopt_long(argc, argv, "ixco:hvV", longopts, &i)) != EOF)
    switch (c) {
      
    case 'i':
      import = 1;
      break;

    case 'x':
      export = 1;
      break;

    case 'c':
      combine = 1;
      break;

    case 'o':
      ofn = optarg;
      break;

    case 'v':
      trace = 1;
      break;
      
    case 'V':
      version = 1;
      return(0);
      break;

    case 'h':
      help = 1;
      return(0);
      break;

    default:
      fprintf(stderr, "sgmcopy error -- unrecognized option %s (use --help to get usage)\n", argv[optind-1]);
      return(1);
    }

  return(0);
}

#undef _sgmcopy_c_
