/******************************************************************************/
/*                                                                            */
/*                                sgmdraw.c                                   */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Guig                                                             Feb, 2003 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2002 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 * $Author: ggravier $
 * $Date: 2003/11/20 15:50:58 $
 * $Revision: 1.6 $
 *
 */

/*
 * Draw samples from a GMM and create an SPro output file.
 */

#define _sgmdraw_c_

#include <audioseg.h>

static char *cvsid = "$Id: sgmdraw.c,v 1.6 2003/11/20 15:50:58 ggravier Exp $";

#define URANDMAX (1.0 - 1.2e-7)
#define NDIV (1 + (2147483647 - 1) / 32)

#define slog(a) (((a) >= 9.22632e-318) ? log(a) : (-1.0e8))

/* ----------------------------------------------- */
/* ----- global variables set by read_args() ----- */
/* ----------------------------------------------- */
unsigned long npts = 10000;    /* number of samples to draw            */
size_t iobs = 10000000;        /* I/O buffer size (in bytes)           */
int trace = 0;                 /* trace level                          */
int help = 0;                  /* print usage                          */
int version = 0;               /* version flag                         */

long seed = 1;                 /* random generator seed                */

/* ---------------------------- */
/* ----- void usage(void) ----- */
/* ---------------------------- */
void usage(void)
{
  fprintf(stdout, "Usage:\n"); 
  fprintf(stdout, "  sgmdraw [options] model ofile\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Synopsis:\n");
  fprintf(stdout, "  Draw samples from a model.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Options:\n");
  fprintf(stdout, "  -S, --buffer-size=n          input buffer size in kbytes (%d)\n", iobs / 1000);
  fprintf(stdout, "  -n, --num-samples=n          number of samples (%ld)\n", npts);
  fprintf(stdout, "\n");
  fprintf(stdout, "  -v, --verbose[=n]            verbose mode\n");
  fprintf(stdout, "  -V, --version                print version number\n");
  fprintf(stdout, "  -h, --help                   this help message\n");
  fprintf(stdout, "\n");
  fflush(stdout);
}

/* ------------------------------------------- */
/* ----- int main(int argc, char** argv) ----- */
/* ------------------------------------------- */
int main(int argc, char** argv) 
{
  char *modfn, *ofn = NULL;
  gmm_t *gmm;
  spfstream_t *f;

  int read_args(int, char **);
  unsigned long draw(gmm_t *, spfstream_t *);

  /* ----- process command line ----- */
  if (read_args(argc, argv))
    return(1);

  if (help) {
    usage();
    return(0);
  }

  if (version) {
    fprintf(stdout, "audioseg %s -- %s\n", VERSION, cvsid);
    return(0);
  }
  
  if (optind < argc)
    modfn = argv[optind++];
  else {
    fprintf(stderr, "sgmdraw error -- no model filename specified (use --help to ge usage)\n");
    return(1);
  }

  if (optind < argc)
    ofn = argv[optind++];
  else {
    fprintf(stderr, "sgmdraw error -- no output filename specified (use --help to ge usage)\n");
    return(1);
  }

  if (optind < argc) {
    fprintf(stderr, "sgmdraw error -- invalid number of arguments (use --help to get usage)\n");
    return(1);
  }

  /* ----- load and initialize the necessary stuff ----- */
  seed = (unsigned)time(NULL);

  if (trace) {
    fprintf(stdout, "loading model from file %s\n", modfn);
    fflush(stdout);
  }

  if ((gmm = gmm_read(modfn)) == NULL) {
    fprintf(stderr, "sgmdraw error -- cannot load model from file %s\n", modfn);
    return(1);
  }

  if ((f = spf_output_stream_open(ofn, get_gmm_dim(gmm), 0, 0, 100.0, NULL, iobs)) == NULL) {
    fprintf(stderr, "sgmdraw error -- cannot open output feature stream %s\n", ofn ? ofn : "stdout");
    gmm_free(gmm);
    return(1);
  }

  /* ----- draw samples ----- */
  if (trace) {
    fprintf(stdout, "drawing %lu samples ...\n", npts);
    fflush(stdout);
  }

  if (draw(gmm, f) != npts) {
    gmm_free(gmm); spf_stream_close(f);
    return(1);
  }

  gmm_free(gmm); 
  spf_stream_close(f);

  return(0);
}

/* ------------------------------------------------------ */
/* ----- unsigned long draw(gmm_t *, spfstream_t *) ----- */
/* ------------------------------------------------------ */
unsigned long draw(gmm_t *gmm, spfstream_t *f)
{
  int k;
  unsigned long i;
  unsigned short j, dim = get_gmm_dim(gmm);
  spf_t *yt;
  float s, u, *m, *v;

  float urand(void);
  float nrand(float, float);

  if ((yt = (spf_t *)malloc(dim * sizeof(spf_t))) == NULL) {
    fprintf(stderr, "sgmdraw error -- cannot allocate memory\n");
    return(0);
  }
  
  /* compute standard deviations from inverse variances once for all */
  for (k = 0; k < get_gmm_num_comp(gmm); k++) {
    v  = get_gmm_comp_ivar(gmm, k);

    for (j = 0; j < dim; j++)
      *(v+j) = (float)(1.0 / sqrt((double)*(v+j)));
  }

  for (i = 0; i < npts; i++) {

    /* choose one component */
    u = urand();
    s = 0.0;
    k = -1;
    while (u > s && k < get_gmm_num_comp(gmm) - 1) { /* k < gmm->nbg - 1 is important since sometimes weights do not sum to 1 exactly! */
      k++;
      s += gmm->w[k];
    }

    /* draw one sample from the selected component */
    m  = get_gmm_comp_mean(gmm, k);
    v  = get_gmm_comp_ivar(gmm, k);

    for (j = 0; j < dim; j++)
      *(yt+j) = nrand(*(m+j), *(v+j));

    if (spf_stream_write(f, yt, 1) != 1) {
      fprintf(stderr, "sgmdraw error -- cannot write frame %lu\n", i);
      break;
    }
  }

  free(yt);
  return(i);  
}

/* ----------------------------- */
/* ----- float urand(void) ----- */
/* ----------------------------- */
float urand(void)
{
  int i;
  long k;
  static long iy = 0;
  static long iv[32];
  static long dum;
  float tmp;

  if (! iy) {
    dum = seed;

    for (i = 32 + 7; i >= 0; i--) {
      k = dum / 127773;
      dum = 16807 * (dum - k * 127773) - k *2836;
      if (dum < 0)
	      dum += 2147483647;
      if (i < 32)
	*(iv+i) = dum;
    }
    iy = *iv;
  }

  k = dum / 127773;
  dum = 16807 * (dum - k * 127773) - k * 2836;
  if (dum < 0)
    dum += 2147483647;

  i = iy / NDIV;
  iy = *(iv+i);
  *(iv+i) = dum;

  tmp =  iy * (1.0 /  2147483647);

  return(tmp > URANDMAX ? URANDMAX : tmp);
}

/* -------------------------------------- */
/* ------ float nrand(float, float) ----- */
/* -------------------------------------- */
float nrand(float m, float s)
{
  static int iset = 0;
  static float gset;
  float fac, rsq, v, v1, v2;

  if (s == 0.0)
    return(m);

  if (! iset) {
    do {
      v1 = 2.0 * urand() - 1.0;
      v2 = 2.0 * urand() - 1.0;
      rsq = v1 * v1 + v2 * v2;
    } while (rsq >= 1.0 || rsq == 0.0);

    fac = sqrt(-2.0 * slog(rsq) / rsq);
    gset = v1 * fac;
    iset = 1;
    v = m + v2 * fac * s;
  } 
  else {
    iset = 0;
    v = m + gset * s;
  }

  return(v);
}

/* --------------------------------------- */
/* ----- int read_args(int, char **) ----- */
/* --------------------------------------- */
/*
 *  -S, --input-bufsize=n        input buffer size in kbytes
 *  -n, --num-samples=n   number of samples to draw
 *
 *  -v, --verbose         verbose mode
 *  -h, --help            help
 */
int read_args(int argc, char **argv)
{
  int c, i;
  char *p;
  struct option longopts[] = {
    {"num-samples", required_argument, NULL, 'n'},
    {"buffer-size", required_argument, NULL, 'S'},
    {"verbose", optional_argument, NULL, 'v'},
    {"help", no_argument, NULL, 'h'},
    {"version", no_argument, NULL, 'V'},
    {0, 0, 0, 0}
  };

  opterr = 0;
  optopt = '!';

  while ((c = getopt_long(argc, argv, "n:S:v::hV", longopts, &i)) != EOF)
    switch (c) {

    case 'n':
      npts = (unsigned long)strtol(optarg, &p, 0);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgmdraw error -- invalid argument %s to option --num-samples (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'S':
      iobs = (size_t)strtol(optarg, &p, 0) * 1000;
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgmdraw error -- invalid argument %s to --buffer-size (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'v':
      trace = 1;
      if (optarg) {
        trace = (int)strtol(optarg, &p, 0);
        if (p == optarg || *p != 0) {
          fprintf(stderr, "sgmdraw error -- invalid argument %s to --verbose (use --help to get usage)\n", optarg);
          return(1);
        }
      }
      break;

    case 'h':
      help = 1;
      return(0);
      break;

    case 'V':
      version = 1;
      return(0);
      break;

    default:
      fprintf(stderr, "sgmdraw error -- unrecognized option %s (use --help to get usage)\n", argv[optind-1]);
      return(1);
    }

  return(0);
}

#undef _sgmdraw_c_
