/******************************************************************************/
/*                                                                            */
/*                                 sgminit.c                                  */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Guig                                                         November 2002 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2002 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 *   $Author: ggravier $
 *   $Date: 2003/11/20 15:50:58 $
 *   $Revision: 1.13 $
 *
 */

/*
 * Gaussian mixture model parameter initialization.
 *
 * The program takes as input either a list of training files or a
 * training script containing a list of data files to use for the
 * paramter estimation, each line following the syntax
 *
 *   /path/to/spro/data/file [path/to/the/corresponding/label/file]
 *
 * where the label file is optionnal. If a label is specified with the
 * --label=s option, only the segments corresponding to that label are
 * used for re-estimation. However, this implies that the label files
 * have been specified in the training script and that the --label
 * option cannot be used without a script.
 *
 * Initialization can be done in one of two ways: uniform or VQ-based.
 *
 * In the uniform initialization, the means are initialized
 * independently in each bin (i.e. dimension) according to the min and
 * max values computed on the training data. 
 *
 * In the VQ-based initialization, a tree-based vector quantization of
 * the training data is first performed before initializing the mean
 * vectors to the tree leaf mean vectors. The tree can be dumped to
 * file using the --save-tree option (see vq. for tree format
 * description).
 *
 * In both cases, weights are initialized to 1/N. Variance is set
 * either to the specified variance floor or to the (possibly floored)
 * global variance of the data. The global variance can be saved to
 * file using the --variance option. This variance file can be
 * directly reused in sgmestim to set a variance floor dependent of
 * the global variance.
 *
 * Trace levels:
 *  1    basic reporting
 *  2    basic tree building report
 *  3    detailed quantization report 
 */

#define _sgminit_c_

#include <audioseg.h>
#include <script.h>
#include <vq.h>
#include <getopt.h>

static char *cvsid = "$Id: sgminit.c,v 1.13 2003/11/20 15:50:58 ggravier Exp $";

/* ----------------------------------------------- */
/* ----- global variables set by read_args() ----- */
/* ----------------------------------------------- */
unsigned short ncomp = 64;     /* number of Gaussians                  */
double globvar = 0;            /* init. variance to global variance    */
double vfloor = 0.0001;        /* variance floor or initial value      */
char *varfn = NULL;            /* (output) variance filename           */
int vq = 0;                    /* quantize                             */
int maxiter = 10;              /* maximum number of iterations         */
double epsilon = 0.0;          /* main loop stopper                    */
int mahalanobis = 0;           /* use Mahalanobis distance             */
char *labname = NULL;          /* extract label name                   */
char *scpfn = NULL;            /* training script file name            */
char *treefn = NULL;           /* save tree to file                    */
long cvtflag = 0;              /* qualifier convertion string          */
size_t iobs = 10000000;        /* I/O buffer size (in bytes)           */
int help = 0;                  /* print usage                          */
int trace = 0;                 /* trace level                          */
int version = 0;               /* version flag                         */

/* ---------------------------- */
/* ----- void usage(void) ----- */
/* ---------------------------- */
void usage(void)
{
  fprintf(stdout, "Usage:\n"); 
  fprintf(stdout, "  sgminit [options] model data ...\n");
  fprintf(stdout, "  sgminit [options] --file-list=script model\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Synopsis:\n");
  fprintf(stdout, "  Gaussian mixture model parameter initialization. If specified, the\n");
  fprintf(stdout, "  input script contains a list of all the training data files,\n");
  fprintf(stdout, "  optionally with the corresponding segmentation files (required if\n");
  fprintf(stdout, "  --label is specified).\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Options:\n");
  fprintf(stdout, "  -n, --num-comp=n             number of Gaussians (%d)\n", ncomp);
  fprintf(stdout, "  -g, --global                 set variance to global variance\n");
  fprintf(stdout, "  -f, --variance-floor=f       variance floor (%.2e)\n", vfloor);
  fprintf(stdout, "  -c, --variance=fn            output global variance to file (don't)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -q, --quantize               VQ-based initialization (no)\n");
  fprintf(stdout, "  -i, --max-iter=i             maximum number of iterations for VQ (%d)\n", maxiter);
  fprintf(stdout, "  -e, --epsilon=f              VQ convergence factor (%.2e)\n", epsilon);
  fprintf(stdout, "  -m, --mahalanobis            use the Mahalanobis distance in the VQ (eucilidian)\n");
  fprintf(stdout, "  -o, --save-tree=fn           save VQ tree to file (don't)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -L, --file-list=fn           input script (none)\n");
  fprintf(stdout, "  -x, --label=s                label (none)\n");
  fprintf(stdout, "  -C, --convert=s              convertion string - ZRADN (none)\n");
  fprintf(stdout, "  -S, --buffer-size=n          input buffer size in kbytes (%d)\n", iobs / 1000);
  fprintf(stdout, "\n");
  fprintf(stdout, "  -v, --verbose[=n]            verbose mode\n");
  fprintf(stdout, "  -V, --version                print version number\n");
  fprintf(stdout, "  -h, --help                   this help message\n");
  fprintf(stdout, "\n");
}

typedef struct {
  unsigned short dim;
  unsigned long nframes;
  double *m;
  double *v;
  double *max;
  double *min;
} spfstat_t;

/* ---------------------------------- */
/* ----- int main(int, char **) ----- */
/* ---------------------------------- */
int main(int argc, char **argv)
{
  script_t *script = NULL;           /* training script                       */
  char *ofn;                         /* output model filename                 */
  seglab_t *label = NULL;
  gmm_t *gmm;

  int read_args(int, char **);
  gmm_t *initialize(script_t *, seglab_t *);

  /* ----- process command line ----- */
  if (read_args(argc, argv))
    return(1);

  if (help) {
    usage();
    return(0);
  }

  if (version) {
    fprintf(stdout, "audioseg %s -- %s\n", VERSION, cvsid);
    return(0);
  }
  
  /* read arguments */
  if (optind < argc)
    ofn = argv[optind++];
  else {
    fprintf(stderr, "sgminit error -- no output filename specified (use --help to get usage)\n");
    return(1);
  }

  if (! scpfn) {
    if (optind >= argc) {
      fprintf(stderr, "sgminit error -- no data filename specified (use --help to get usage)\n");
      return(1);
    }
    if ((script = make_script(argc - optind, argv + optind)) == NULL) {
      fprintf(stderr, "sgminit error -- cannot create script from arguments\n");
      return(1);
    }
  }
  else if ((script = load_script(scpfn, 0)) == NULL) {
    fprintf(stderr, "sgminit error -- cannot load script from file %s\n", scpfn);
    return(1);
  }
  
  /* ----- load segmentations and create label if necessary ----- */
  if (labname) {
    if ((label = str_to_label(labname)) == NULL) {
      fprintf(stderr, "sgminit error -- cannot create reference label\n");
      free_script(script);
      return(1);
    }

    if (load_script_seg(script) == NULL) {
      fprintf(stderr, "sgminit error -- cannot load segmentation\n");
      free_script(script); seg_label_free(label);
      return(1);
    }
  }

  /* ----- do the initialization job ----- */
  if (trace) {
    fprintf(stdout, "script    = %s (%lu files)\n", scpfn ? scpfn : "<nil>", script->nfiles);
    fprintf(stdout, "label     = %s\n", labname ? labname : "<nil>");
    fprintf(stdout, "quantize  = %s\n", vq ? "yes" : "no");

  }
  if ((gmm = initialize(script, label)) == NULL) {
    free_script(script); seg_label_free(label);
    return(1);
  }

  /* ----- write model to output file ----- */
  if (trace) {
    fprintf(stdout, "saving model to file %s\n", ofn);
    fflush(stdout);
  }

  if (gmm_write(ofn, gmm) != 0) {
    fprintf(stderr, "sgminit error -- cannot write model to output file %s", ofn);
    free_script(script); seg_label_free(label); gmm_free(gmm);
    return(1);
  }

  gmm_free(gmm);
  seg_label_free(label);
  free_script(script);

  return(0);
}

/* ----------------------------------------------------- */
/* ----- gmm_t *initialize(script_t *, seglab_t *) ----- */
/* ----------------------------------------------------- */
/*
 * Top-level initialization procedure.
 */
gmm_t *initialize(script_t *script, seglab_t *label)
{
  FILE *f;
  spfstat_t *stats;
  gmm_t *gmm;
  double *w = NULL;
  unsigned short j, dim;
  unsigned long nframes;
  vqtree_t *tree = NULL;

  spfstat_t *get_stats(script_t *, seglab_t *);
  void spf_stat_free(spfstat_t *);
  int build_tree(vqtree_t *, double *, script_t *, seglab_t *);
  gmm_t *create_model(spfstat_t *, vqtree_t *);

  /* ----- compute basic statistics ----- */
  if (trace) {
    fprintf(stdout, "computing global statistics");
    fflush(stdout);
  }
  if ((stats = get_stats(script, label)) == NULL)
    return(NULL);

  dim = stats->dim;

  if ((nframes = stats->nframes) == 0) {
    fprintf(stderr, "sgminit error -- empty training set\n");
    return(NULL);
  }

  /* ----- create model ----- */
  if (trace) {
    fprintf(stdout, "  dim=%u  nframes=%lu\n", dim, nframes);
    fflush(stdout);
  }

  if (vq) {

    if (trace) {
      fprintf(stdout, "clustering data\n");
      fflush(stdout);
    }
    
    if (mahalanobis) {
      if ((w = (double *)malloc(dim * sizeof(double))) == NULL) {
	spf_stat_free(stats);
	fprintf(stderr, "sgminit error -- cannot allocate memory\n");
	return(NULL);
      }

      for (j = 0; j < dim; j++)
	*(w+j) = 1.0 / *(stats->v+j);
    }
    
    if ((tree = init_tree(ncomp, dim, stats->m, stats->v, nframes)) == NULL) {
      fprintf(stderr, "sgminit error -- cannot allocate memory\n");
      spf_stat_free(stats); if (w) free(w);
      return(NULL);
    }

    if (build_tree(tree, w, script, label) != 0) {
      spf_stat_free(stats); free_tree(tree); if (w) free(w);
      return(NULL);
    }

    if (w)
      free(w);

    if (treefn) {
      if ((f = fopen(treefn, "wt")) == NULL)
	fprintf(stderr, ">>>>> WARNING >>>>>> cannot open tree file %s (tree not saved)\n", treefn);
      else {
	print_tree(f, tree);
	fclose(f);
      }
    }
  }

  if ((gmm = create_model(stats, tree)) == NULL) {
    spf_stat_free(stats); 
    if (tree) free_tree(tree);
    return(NULL);
  }

  if (varfn) {
    if ((f = fopen(varfn, "wt")) == NULL)
      fprintf(stderr, ">>>>> WARNING >>>>>> cannot open variance file %s (variance not saved)\n", varfn);
    else {
      for (j = 0; j < dim; j++)
	fprintf(f, j ? " %lf" : "%lf", stats->v[j]);
      fclose(f);
    }
  }
  
  spf_stat_free(stats);
  free_tree(tree);

  return(gmm);
}

/* ------------------------------------------------------------------------ */
/* ----- int build_tree(vqtree_t *, double *, script_t *, seglab_t *) ----- */
/* ------------------------------------------------------------------------ */
/*
 * Grow the tree until the desired number of leaves is reached. Return
 * 0 if everything's ok.
 */
int build_tree(vqtree_t *tree, double *w, script_t *script, seglab_t *label)
{
  unsigned short *idx;
  unsigned long i;
  unsigned short nleaves, n, offset;
  int m, it;
  double d, dmem = 0.0;
  vqacc_t *acc;

  int partition(vqtree_t *, unsigned short *, vqacc_t *, unsigned short, double *, script_t *, seglab_t *, double *);

  frexp(ncomp, &m);

  /* 
     Allocate a temporary structure to store the node for each data point. The 
     total number of data points is actually the number of points of the root 
     node.
  */
  if ((idx = (unsigned short *)malloc(tree->node[0].npts * sizeof(unsigned short))) == NULL) {
    fprintf(stderr, "sgminit error -- cannot allocate memory\n");
    return(1);
  }
  for (i = 0; i < tree->node[0].npts; i++)
    *(idx+i) = 0;

  /*
    Allocate memory for leaf accumulators
  */
  if ((acc = vq_acc_alloc(ncomp, tree->dim)) == NULL) {
    fprintf(stderr, "sgminit error -- cannot allocate memory\n");
    free(idx);
    return(1);
  }

  /*
    Run clustering until number of leaves is reached.
  */
  for (n = 1; n < m; n++) {
    nleaves = 1 << n;
    offset = (1 << n) - 1;

    split(tree);
    if (trace > 1) {
      fprintf(stdout, "   %d leaves -> %d [nnodes=%d]\n", nleaves / 2, nleaves, tree->nnodes);
      fflush(stdout);
    }

    for (it = 0; it < maxiter; it++) {
      vq_acc_reset(acc, nleaves, tree->dim);
      for (i = 0; i < nleaves; i++)
	tree->node[offset+i].npts = 0;

      if (trace > 1) {
	fprintf(stdout, "iteration %d: ", it+1);
	fflush(stdout);
      }

      if (partition(tree, idx, acc, offset, w, script, label, &d) != 0) {
	vq_acc_free(acc, ncomp);
	return(1);
      }

      if (trace > 1) {
	fprintf(stdout, "    iteration %d:  dist=%.12e", it + 1, d / (double)(tree->node[0].npts));
	if (it)
	  fprintf(stdout, "  diff=%.12e  increase=%.12e", (dmem - d) / (double)(tree->node[0].npts), fabs((dmem - d) / dmem));
	fprintf(stdout, "\n");
	fflush(stdout);
      }

      if ( (it > 0) && (fabs((dmem - d) / dmem) < epsilon) )
	break;

      if (update(tree, acc, offset) != 0) {
	vq_acc_free(acc, ncomp);
	return(1);
      }

      dmem = d;
    }

    /* 
       WATCH OUT: the tree structure is corrupted at that point so
       that the childs of a leaf are actually the brothers (i.e. the
       leaf itself and its sister) rather than null. This is a trick
       for partition() to work properly over the iterations but we
       gotta fix that so that the next split() works correctly...
    */
    for (i = 0; i < nleaves; i++)
      tree->node[offset+i].child[0] = tree->node[offset+i].child[1] = VQ_NULL_ID;

  }

  /* that's it, man! */
  free(idx);
  vq_acc_free(acc, ncomp);

  return(0);
}

/* -------------------------------------------------------- */
/* ----- gmm_t *create_model(spfstat_t *, vqtree_t *) ----- */
/* -------------------------------------------------------- */
/*
 * Create model from tree or from min and max vectors.
 */
gmm_t *create_model(spfstat_t *stats, vqtree_t *tree)
{
  gmm_t *gmm;
  unsigned short i, j, nid;
  float *p;
  double x;
  vqnode_t *node;
  unsigned short dim = stats->dim;

  if ((gmm = gmm_alloc(ncomp, dim)) == NULL) {
    fprintf(stderr, "sgminit error -- cannot allocate memory\n");
    return(NULL);
  }

  /* set mean vectors */
  if (tree) {
    i = 0;

    for (nid = 0; nid < tree->nnodes; nid++) {
      node = tree->node + nid;

      if (is_leaf(node)) {
	p = get_gmm_comp_mean(gmm, i);

	for (j = 0; j < dim; j++)
	  *(p+j) = (float)*(node->m+j);

	i++;
      }
    }
  }
  else
    for (j = 0; j < dim; j++) {

      x = (stats->max[j] - stats->min[j]) / (double)ncomp;
      gmm->m[0][j] = stats->min[j] + 0.5 * x;

      for (i = 1; i < ncomp; i++)
	gmm->m[i][j] = gmm->m[i-1][j] + x;
    }
  
  /* set variances and weights */
  x = 1.0 / (double)ncomp;
  for (i = 0; i < ncomp; i++) {
    gmm->w[i] = (float)x;
    p = get_gmm_comp_ivar(gmm, i);
    
    if (globvar)
      for (j = 0; j < dim; j++)
	*(p+j) = (stats->v[j] > vfloor) ? ((float)(1.0 / stats->v[j])) : (1.0 / vfloor);
    else
      for (j = 0; j < dim; j++)
	*(p+j) = vfloor;
  }

  return(gmm);
}

/* ---------------------------------------------------------------------------------- */
/* ----- int partition(vqtree_t *, unsigned short *, vqacc_t *, unsigned short, ----- */
/* ----                double *, script_t *, seglab_t *, double *)              ----- */
/* ---------------------------------------------------------------------------------- */
/*
 * Partition the data into the tree leaves. For each sample in the
 * training set, find out the best cluster between the sisters of its
 * current cluster and modify the index table (idx)
 * accordingly. Accumulate sufficient statistics for each leaf and
 * compute the overall distortion.
 */
int partition(vqtree_t *tree, unsigned short *idx, vqacc_t *acc, unsigned short offset, 
	      double *w, script_t *script, seglab_t *label, double *dtot)
{
  asseg_t *curseg;
  spfstream_t *f;
  spf_t *yt;
  float st, et;
  unsigned long i, n, t, sn, en, nframes;
  unsigned short j, ac, nc, c, dim = tree->dim;
  double x, *m, *v, d, segd = 0;
  
  asseg_t *get_next_seg(float *, float *, asseg_t *, seglab_t *, int);

  n = 0;
  *dtot = 0.0;

  /* ----- iterate on each files ----- */
  for (i = 0; i < script->nfiles; i++) {
    
    curseg = get_next_seg(&st, &et, label ? script->seg[i] : NULL, label, 1);

    if (st == ASEG_NULL_TIME)
      continue;

    /* open feature stream */
    if ((f = spf_input_stream_open(script->datfn[i], cvtflag, iobs)) == NULL) {
      fprintf(stderr, "sgminit error -- cannot open feature stream %s\n", script->datfn[i]);
      return(1);
    }

    /* loop on all segments */
    while (st != ASEG_NULL_TIME) {
      
      /* get initial start and possibly end frame number and seek in
	 input stream */
      sn = (unsigned long)(st * spf_stream_rate(f));
      if (et != ASEG_NULL_TIME) {
	en = (unsigned long)(et * spf_stream_rate(f));
	if (en == sn)
	  en += 1;
      }
      else
	en = 0;

      if (spf_stream_seek(f, sn, SEEK_SET) != 0) {
	fprintf(stderr, "sgminit error -- cannot seek frame %lu at t=%.f in stream %s\n", sn, st, script->datfn[i]);
	spf_stream_close(f);
	return(1);
      }

      segd = 0.0;
      nframes = 0;
      t = sn;

      /* do partitioning for current segment */
      while ((yt = get_next_spf_frame(f)) != NULL) {

	c = *(idx+n); /* parent codeword */
	nc = quantize(yt, w, dim, tree->node + tree->node[c].child[0], tree->node + tree->node[c].child[1], &d);
	nc = tree->node[c].child[nc]; /* new codeword */
	ac = nc - offset; /* accumulator index */
	
	*(idx+n) = nc;
	tree->node[nc].npts += 1;
	m = acc[ac].m;
	v = acc[ac].v;
	for (j = 0; j < dim; j++) {
	  x = *(yt+j);
	  *(m+j) += x;
	  *(v+j) += x * x;
	}

        nframes++;
	segd += d;
	n++;

	t++;
	if (t == en)
	  break;
      }

      if (trace > 2) {
	fprintf(stdout, "   file=%s  st=%.2f  et=%.2f  sn=%ld  en=%ld  nframes=%lu  dist=%e\n", script->datfn[i], st, et, sn, en, nframes, segd / nframes);
	fflush(stdout);
      }

      *dtot += segd;

      /* switch to next segment */
      curseg = get_next_seg(&st, &et, curseg, label, 0);
    }

    spf_stream_close(f);
  }

  return(0);
}

/* -------------------------------------------------------- */
/* ----- spfstat_t *get_stats(script_t *, seglab_t *) ----- */
/* -------------------------------------------------------- */
/*
 * Compute basic statistics on the features, namely global mean and
 * variance, highest and lowest value.
 */
spfstat_t *get_stats(script_t *script, seglab_t *label)
{
  spfstream_t *f;
  spf_t *yt;
  unsigned short dim = 0;
  spfstat_t *stats = NULL;
  asseg_t *curseg;
  unsigned long i, t, sn, en, nframes;
  float st, et;
  unsigned short j;
  double x;

  spfstat_t *spf_stat_alloc(unsigned short);
  void spf_stat_free(spfstat_t *);
  asseg_t *get_next_seg(float *, float *, asseg_t *, seglab_t *, int);

  /* ----- iterate on each files ----- */
  for (i = 0; i < script->nfiles; i++) {
    
    curseg = get_next_seg(&st, &et, label ? script->seg[i] : NULL, label, 1);
    
    if (st == ASEG_NULL_TIME)
      continue;

    /* open feature stream */
    if ((f = spf_input_stream_open(script->datfn[i], cvtflag, iobs)) == NULL) {
      fprintf(stderr, "sgminit error -- cannot open feature stream %s\n", script->datfn[i]);
      return(NULL);
    }

    /* initialize the accumulators */
    if (dim == 0) {
      dim = spf_stream_dim(f);
      
      if ((stats = spf_stat_alloc(dim)) == NULL) {
	fprintf(stderr, "sgminit error -- cannot allocate memory\n");
	spf_stream_close(f);
	return(NULL);
      }

      stats->nframes = 0;
      for (j = 0; j < dim; j++) {
	stats->m[j] = 0.0;
	stats->v[j] = 0.0;
	stats->min[j] = FLT_MAX;
	stats->max[j] = -FLT_MAX;
      }
    }
    else if (spf_stream_dim(f) != dim) { /* check it's a valid file */
      fprintf(stderr,  
	      "sgminit error -- incompatible data (%d) and model (%d) dimensions in file %s\n", 
	      spf_stream_dim(f), dim, script->datfn[i]);
      spf_stream_close(f); spf_stat_free(stats);
      return(NULL);
    }

    /* loop on all segments */
    while (st != ASEG_NULL_TIME) {
      
      /* get initial start and possibly end frame number and seek in
	 input stream */
      sn = (unsigned long)(st * spf_stream_rate(f)); 
      if (et != ASEG_NULL_TIME) {
	en = (unsigned long)(et * spf_stream_rate(f));
	if (en == sn)
	  en += 1;
      }
      else
	en = 0;

      if (spf_stream_seek(f, sn, SEEK_SET) != 0) {
	fprintf(stderr, "sgminit error -- cannot go to frame %lu at t=%.f in file %s\n", sn, st, script->datfn[i]);
	spf_stream_close(f);
	return(NULL);
      }

      nframes = 0;
      t = sn;
      
      /* accumulate stats for current segment */
      while ((yt = get_next_spf_frame(f)) != NULL) {
	
	for (j = 0; j < dim; j++) {
	  x = (double)*(yt+j);

	  *(stats->m+j) += x;
	  *(stats->v+j) += x * x;

	  if (x > *(stats->max+j))
	    *(stats->max+j) = x;

	  if (x < *(stats->min+j))
	    *(stats->min+j) = x;
	}

	nframes++;
	t++;
	if (t == en)
	  break;
      }
      stats->nframes += nframes;
   
      curseg = get_next_seg(&st, &et, curseg, label, 0);
    }

    spf_stream_close(f);
  }

  /* ----- compute actual mean and variance ----- */
  for (j = 0; j < dim; j++) {
    x = *(stats->m+j) / stats->nframes;
    *(stats->m+j) = x;
    *(stats->v+j) = *(stats->v+j) / stats->nframes - x * x;
  }

  return(stats);
}

/* ------------------------------------------------------------------------------- */
/* ----- asseg_t *get_next_seg(float *, float *, asseg_t *, seglab_t *, int) ----- */
/* ------------------------------------------------------------------------------- */
/*
 * Get next segment start and end times. If no more segments, set st
 * to ASEG_NULL_TIME. If end of segment is unkown (i.e. it matches the
 * end of the input feature stream) et is set to ASEG_NULL_TIME. In
 * any other cases, st and et are set to the segment start and end
 * times (in seconds). If a segmentation is supplied, move to the next
 * segment in the segmentation, searching for the specified label if
 * there's one. In this case, the function returns a pointer to the
 * segment found to be used in successive calls to get_next_seg().
 *
 * The function is initialized when first invoked with a non null
 * reset flag.
 */
asseg_t *get_next_seg(float *st, float *et, asseg_t *seg, seglab_t *label, int reset)
{
  asseg_t *p = NULL;


  if (seg) {
    p = (reset) ? seg : seg->next;
    if (! p) {
      *st = *et = ASEG_NULL_TIME;
      return(NULL);
    }
    while (seg_label_compare(p->label, label) != 0)
      if ((p = p->next) == NULL) {
	*st = *et = ASEG_NULL_TIME;
	return(NULL);
      }
    *st = p->st;
    *et = p->et;
  }
  else {
    *st = (reset) ? 0.0 : ASEG_NULL_TIME;
    *et = ASEG_NULL_TIME;
  }

  return(p);
}

/* ---------------------------------------------------- */
/* ---- spfstat_t *spf_stat_alloc(unsigned short) ----- */
/* ---------------------------------------------------- */
/*
 * Allocate memory for statistics.
 */
spfstat_t *spf_stat_alloc(unsigned short dim)
{
  spfstat_t *p;
  void spf_stat_free(spfstat_t *);
  
  p = (spfstat_t *)malloc(sizeof(spfstat_t));

  if (p) {
    p->nframes = 0;
    p->dim = dim;
    p->m = (double *)malloc(dim * sizeof(double));
    p->v = (double *)malloc(dim * sizeof(double));
    p->min = (double *)malloc(dim * sizeof(double));
    p->max = (double *)malloc(dim * sizeof(double));

    if (p->m == NULL || p->v == NULL || p->min == NULL || p->max == NULL) {
      spf_stat_free(p);
      return(NULL);
    }
  }

  return(p);
}

/* ------------------------------------------- */
/* ----- void spf_stat_free(spfstat_t *) ----- */
/* ------------------------------------------- */
/*
 * Free memory allocated for statistics.
 */
void spf_stat_free(spfstat_t *p)
{
  if (p) {
    if (p->m)
      free(p->m);
    if (p->v)
      free(p->v);
    if (p->min)
      free(p->min);
    if (p->max)
      free(p->max);
    free(p);
  }
}

/* --------------------------------------- */
/* ----- int read_args(int, char **) ----- */
/* --------------------------------------- */
/*
 *  -n, --num-comp=n             number of Gaussians (64)
 *  -g, --global                 initalize variance to global variance
 *  -f, --varfloor=f             variance floor
 *  -q, --quantize               VQ-based initialization (no)
 *  -i, --max-iter               maximum number of iterations for VQ (10)
 *  -m, --mahalanobis            use the Mahalanobis distance in the VQ (eucilidian)
 *  -o, --save-tree=fn           save VQ tree to file (don't)
 *  -L, --file-list=fn           input script (none)
 *  -l, --label=s                label (none)
 *  -C, --convert=s              convertion string - ADNZ (none)
 *  -S, --input-bufsize=n        input buffer size in kbytes
 *  -v, --verbose[=n]            verbose mode
 *  -V, --version                print version number
 *  -h, --help                   this help message
 */
int read_args(int argc, char **argv)
{
  int c, i;
  char *p;
  struct option longopts[] = {
    {"num-comp", required_argument, NULL, 'n'},
    {"max-iter", required_argument, NULL, 'i'},
    {"global", no_argument, NULL, 'g'},
    {"variance-floor", required_argument, NULL, 'f'},
    {"variance", required_argument, NULL, 'c'},
    {"quantize", no_argument, NULL, 'q'},
    {"mahalanobis", no_argument, NULL, 'm'},
    {"save-tree", required_argument, NULL, 'o'},
    {"file-list", required_argument, NULL, 'L'},
    {"label", required_argument, NULL, 'x'},
    {"convert", required_argument, NULL, 'C'},
    {"buffer-size", required_argument, NULL, 'S'},
    {"verbose", optional_argument, NULL, 'v'},
    {"version", no_argument, NULL, 'V'},
    {"help", no_argument, NULL, 'h'},
    {0, 0, 0, 0}
  };

  opterr = 0;
  optopt = '!';

  while ((c = getopt_long(argc, argv, "i:n:f:c:gqmo:L:x:C:S:v::Vh", longopts, &i)) != EOF)
    switch (c) {
      
    case 'n':
      ncomp = (unsigned short)strtol(optarg, &p, 0);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgminit error -- invalid argument %s to --num-comp (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'f':
      vfloor = (double)strtod(optarg, &p);
      if (p == optarg  || *p != 0) {
	fprintf(stderr, "sgminit error -- invalid argument %s to --varfloor (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'c':
      varfn = optarg;
      break;

    case 'g':
      globvar = 1;
      break;

    case 'q':
      vq = 1;
      break;
      
    case 'i':
      maxiter = (int)strtol(optarg, &p, 0);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgminit error -- invalid argument %s to --max-iter (use --help to get usage)\n", optarg);
	return(1);
      }
      break; 

    case 'm':
      mahalanobis = 1;
      break;

    case 'o':
      treefn = optarg;
      break;

    case 'L':
      scpfn = optarg;
      break;

    case 'x':
      labname = optarg;
      break;

    case 'C': 
      cvtflag = sp_str_to_flag(optarg);      
      break;

    case 'S':
      iobs = (size_t)strtol(optarg, &p, 0) * 1000;
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgminit error -- invalid argument %s to --buffer-size (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'v':
      trace = 1;
      if (optarg) {
	trace = (int)strtol(optarg, &p, 0);
	if (p == optarg || *p != 0) {
	  fprintf(stderr, "sgminit error -- invalid argument %s to --verbose (use --help to get usage)\n", optarg);
	  return(1);
	}
      }
      break;

    case 'V':
      version = 1;
      break;

    case 'h':
      help = 1;
      return(0);
      break;

    default:
      fprintf(stderr, "sgminit error -- unrecognized option %s (use --help to get usage)\n", argv[optind-1]);
      return(1);
    }

  /* argument sanity check */
  if (vfloor < 0.0) {
    fprintf(stderr, "sgminit error -- invalid variance floor value %e (not positive)\n", vfloor);
    return(1);
  }

  return(0);
}

#undef _sgminit_c_
