/******************************************************************************/
/*                                                                            */
/*                                sgmlike.c                                   */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Guig                                                             Aug, 2002 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2002 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 * $Author: ggravier $
 * $Date: 2003/11/20 15:50:58 $
 * $Revision: 1.14 $
 *
 */

/*
 * Likelihood computation.
 *
 * The log-likelihood computation program takes as input a model or a
 * list of models and one or several datafiles. The list of input
 * models is a text file containing the paths to models, one per
 * line. Alternately, the program can load a segmentation file and
 * score on the basis of the given segmentation.

 * The program outputs the scores foreach specified model and for each
 * segment of each datafile. For each segment of each input file, a
 * line whith the following syntax is written
 *
 *  filename start_time end_time nframes name nmodels llk[0] .... llk[n]
 *
 * where llk[i] is the segment log-likelihood with model j, start_time
 * and end_time the segment start and end times (-1 for end of stream)
 * in seconds wrt the beginning of the file, nframes the corresponding
 * number of frames, name is the segment name (or - if none was
 * specified) and nmodels is the number of models. By default, the
 * output is made to stdout unless otherwised specified with the
 * --output=fn option.
 *
 * The program can also compute posterior probabilities rather than
 * log-likelihoods. If a set of model M_i is loaded, the posterior
 * probability of a segment y_[t1,t2] is given foreach model by
 * P[M_i|y_[t1,t2] = \frac{p[y_[t1,t2]|M_i}{\sum_j p[y_[t1,t2]|M_j]} .
 *
 * The program can take its datafile input from stdin by specifying -
 * as the data filename.
 */

#define _sgmlike_c_

#include <audioseg.h>
#include <script.h>
#include <getopt.h>

static char *cvsid = "$Id: sgmlike.c,v 1.14 2003/11/20 15:50:58 ggravier Exp $";

# define MAX_N_MODELS 256

/* ----------------------------------------------- */
/* ----- global variables set by read_args() ----- */
/* ----------------------------------------------- */
unsigned short ncomp = 64;     /* number of Gaussians                  */
char *listfn = NULL;           /* model list filename                  */
char *scpfn = NULL;            /* script filename                      */
float frate = 100;             /* frame rate                           */
int use_seg = 0;               /* use existing segmentation            */
char *segfn = NULL;            /* segmentation filename                */
char *labname = NULL;          /* extract label name                   */
float length = 0;              /* segment length (in seconds)          */
float shift = 0;               /* segment shift (in seconds)           */
float startt = 0.0;            /* start time                           */
float endt = ASEG_NULL_TIME;   /* end time                             */
unsigned short nbest = 0;      /* N-best Gaussian scoring              */
int normalize = 0;             /* normalize llk by segment length      */
int post = 0;                  /* output posterior probabilities       */
char *ofn = NULL;              /* output filename                      */
long cvtflag = 0;              /* qualifier convertion string          */
size_t iobs = 10000000;        /* I/O buffer size (in bytes)           */
int add = 0;                   /* add segment scores when scoring segs */
int help = 0;                  /* print usage                          */
int trace = 0;                 /* trace level                          */
int version = 0;               /* print version information            */

/* ---------------------------- */
/* ----- void usage(void) ----- */
/* ---------------------------- */
void usage(void)
{
  fprintf(stdout, "Usage:\n"); 
  fprintf(stdout, "  sgmlike [options] model data ...\n");
  fprintf(stdout, "  sgmlike [options] --list=fn data ...\n");
  fprintf(stdout, "  sgmlike [options] --file-list=fn model\n");
  fprintf(stdout, "  sgmlike [options] --list=fn --file-list=fn\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Synopsis:\n");
  fprintf(stdout, "  Score segments in data files with the specified model(s). If\n");
  fprintf(stdout, "  --segmentation is specified, sgmlike scores each segment in the\n");
  fprintf(stdout, "  specified segmentation file(s). The segmentation is taken from the\n");
  fprintf(stdout, "  file list if it is specified, otherwise the program will load the\n");
  fprintf(stdout, "  segmentation from the file specified as argument to --segmentation\n");
  fprintf(stdout, "  option (which means only one file can be specified in this case).\n");
  fprintf(stdout, "  Fixed length segmentation can be specified with the --length and\n");
  fprintf(stdout, "  --shift options while the --time option can be used to specify a\n");
  fprintf(stdout, "  particular portion of the input file(s) to be scored.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Options:\n");
  fprintf(stdout, "  -s, --segmentation[=fn]      use segmentation for scoring (don't)\n");
  fprintf(stdout, "  -x, --label=s                score segments with specified name (none)\n");
  fprintf(stdout, "  -l, --length=n               segment length in seconds (all file)\n");
  fprintf(stdout, "  -d, --shift=n                segment shift in seconds (segment length)\n");
  fprintf(stdout, "  -t, --time=[st][:et]         score between start and end times (all file)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -a, --add                    add segment scores (don't)\n");
  fprintf(stdout, "  -b, --n-best=n               N-best Gaussian scoring (no)\n");
  fprintf(stdout, "  -n, --normalize              normalize score by segment length (don't)\n");
  fprintf(stdout, "  -p, --posterior              output posterior probabilities (log-likelihoods)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -L, --file-list=fn           input script (none)\n");
  fprintf(stdout, "  -f, --list=fn                set model list (none)\n");
  fprintf(stdout, "  -C, --convert=s              convertion string - ADNZ (none)\n");
  fprintf(stdout, "  -S, --buffer-size=n          input buffer size in kbytes (%d)\n", iobs / 1000);  
  fprintf(stdout, "\n");
  fprintf(stdout, "  -o, --output=fn              output file (stdout)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -v, --verbose[=n]            verbose mode\n");
  fprintf(stdout, "  -V, --version                print version number\n");
  fprintf(stdout, "  -h, --help                   this help message\n");
  fprintf(stdout, "\n");
}

/* ----- local structures ----- */
typedef struct {
  unsigned short nmodels;
  gmm_t **gmm;
} gmmlist_t;

/* ---------------------------------- */
/* ----- int main(int, char **) ----- */
/* ---------------------------------- */
int main(int argc, char** argv)
{
  FILE *ostream;
  gmmlist_t *mlist = NULL;
  script_t *script;
  seglab_t *label = NULL;
  unsigned long i;

  int read_args(int, char **);
  gmmlist_t *gmm_list_read(const char *, int);
  void gmm_list_free(gmmlist_t *);
  int score_file(const char *, const char *, seglab_t *, gmmlist_t *, FILE *);


  /* ----- process command line ----- */
  if (read_args(argc, argv))
    return(1);

  if (help) { 
    usage();
    return(0);
  }

  if (version) {
    fprintf(stdout, "audioseg %s -- %s\n", VERSION, cvsid);
    return(0);
  }
  
  /* read arguments */
  if (! listfn) {
    if (optind >= argc) {
      fprintf(stderr, "sgmlike error -- no model name specified (use --help to get usage)\n");
      return(1);
    }
    if ((mlist = gmm_list_read(argv[optind++], 0)) == NULL) {
      fprintf(stderr, "sgmlike error -- cannot create model list from arguments\n");
      return(1);
    }
  }
  else if ((mlist = gmm_list_read(listfn, 1)) == NULL) {
    fprintf(stderr, "sgmlike error -- cannot load model list from file %s\n", listfn);
    return(1);
  }
  
  if (! scpfn) {
    if (optind >= argc) {
      fprintf(stderr, "sgmlike error -- no data filename specified (use --help to get usage)\n");
      gmm_list_free(mlist);
      return(1);
    }
    if ((script = make_script(argc - optind, argv + optind)) == NULL) {
      gmm_list_free(mlist);
      fprintf(stderr, "sgmlike error -- cannot create script from arguments\n");
      return(1);
    }
    if (segfn) {
      if (script->nfiles != 1) {
	gmm_list_free(mlist); free_script(script);
	fprintf(stderr, "sgmlike error -- too many parameters for option -s (use -help to get usage)\n");
	return(1);
      }
      if ((script->segfn[0] = strdup(segfn)) == NULL) {
	gmm_list_free(mlist); free_script(script);
	fprintf(stderr, "sgmlike error -- cannot allocate memory\n");
	return(1);
      }
    }
  }
  else if ((script = load_script(scpfn, 0)) == NULL) {
    gmm_list_free(mlist);
    fprintf(stderr, "sgmlike error -- cannot load script from file %s\n", scpfn);
    return(1);
  }

  /* ---- open output file if necessary ----- */
  if (ofn) {
    if ((ostream = fopen(ofn, "w")) == NULL) {
      fprintf(stderr, "sgmlike error -- cannot open output file %s\n", ofn);
      gmm_list_free(mlist); free_script(script);
      return(1);
    }
  }
  else
    ostream = stdout;

  /* ----- create label if necessary ----- */
  if (labname)
    if ((label = str_to_label(labname)) == NULL) {
      fprintf(stderr, "sgmlike error -- cannot create reference label\n");
      gmm_list_free(mlist); free_script(script); if (ofn) fclose(ostream);
      return(1);
    }
  

  /* ----- loop on all input files ----- */
  for (i = 0; i < script->nfiles; i++) {

    if (trace) {
      fprintf(stdout, "file=%s (%s)\n", script->datfn[i], labname ? labname : "nil");
      fflush(stdout);
    }
    
    if (score_file(script->datfn[i], use_seg ? script->segfn[i] : NULL, label, mlist, ostream) == 0) {
      fprintf(stderr, ">>>>> WARNING >>>>> cannot score input file %s (ignoring)\n", script->datfn[i]);
      continue;
    }
  }

  if (ofn) 
    fclose(ostream);
  gmm_list_free(mlist); 
  free_script(script);
  seg_label_free(label);

  return(0);
}

/* ------------------------------------------------------------------------------------------------- */
/* ----- unsigned long score_file(const char *, const char *, seglab_t *, gmmlist_t *, FILE *) ----- */
/* ------------------------------------------------------------------------------------------------- */
/*
 * Score content of a file, loading any corresponding segmentation of
 * required. Return the number of segments scored.
 */
unsigned long score_file(const char *ifn, const char *sfn, seglab_t *label, gmmlist_t *mlist, FILE *ostream)
{
  spfstream_t *f;
  asseg_t *seg, *curseg;
  unsigned short i, dim;
  unsigned long ntot = 0, nframes, nsegs = 0;
  double Z, S, logp[MAX_N_MODELS], score[MAX_N_MODELS];
  float st, et;

  asseg_t *get_next_seg(float *, float *, asseg_t *, seglab_t *, int);
  unsigned long score_segment(spfstream_t *, float, float, gmmlist_t *, double *);

  dim = get_gmm_dim(mlist->gmm[0]);

  /* load segmentation if necessary */
  seg = (asseg_t *)NULL;
  if (sfn)
    if ((seg = seg_read(sfn)) == NULL) {
      fprintf(stderr,"sgmlike error -- cannot load segmenation file %s\n", sfn);
      return(0);
    }

  /* open feature stream */
  if ((f = spf_input_stream_open(ifn, cvtflag, iobs)) == NULL) {
    fprintf(stderr, "sgmlike error -- cannot open feature stream %s\n", ifn ? ifn : "stdin");
    seg_list_free(seg);
    return(1);
  }

  /* check out model and file compatibility */
  if (spf_stream_dim(f) != dim) {
    fprintf(stderr, "sgmlike error -- model dimension %d differs from frame dimension %d\n", dim, spf_stream_dim(f));
    seg_list_free(seg); spf_stream_close(f);
    return(0);
  }

  /* run scoring */
  if (add)
    for (i = 0; i < mlist->nmodels; i++)
      score[i] = 0.0;

  curseg = get_next_seg(&st, &et, seg, label, 1);
  
  while ((nframes = score_segment(f, st, et, mlist, logp)) > 0) {

    if (add) {
      ntot += nframes;
      for (i = 0; i < mlist->nmodels; i++)
	*(score+i) += *(logp+i);
    }
    else {

      if (normalize || post) {
	Z = 1.0 / (double)nframes;
	S = 0.0;

	for (i = 0; i < mlist->nmodels; i++) {
	  *(logp+i) *= Z;
	  if (post)
	    S +=  exp(*(logp+i));
	}
	if (post) {
	  Z = 1.0 / S;
	  for (i = 0; i < mlist->nmodels; i++)
	    *(logp+i) = exp(*(logp+i)) * Z;
	}
      }

      /* output format is:  filename start_time nframes name nmodels llk[0] .... llk[n] */
      fprintf(ostream, "%s %.2f %.2f %ld", ifn, st, et, nframes);
      if (curseg)
	for (i = 0; i < curseg->label->nlabels; i++)
	  fprintf(ostream, i ? "+%s" : " %s", curseg->label->name[i]);
      else
	fprintf(ostream, " -");
      fprintf(ostream, " %d", mlist->nmodels);
      for (i = 0; i < mlist->nmodels; i++)
	fprintf(ostream, " %-10.4f", *(logp+i));
      fprintf(ostream, "\n");
    }

    nsegs++;
    curseg = get_next_seg(&st, &et, curseg, label, 0);
  }

  /* output total stuff if add was set */
  if (add) {
    if (normalize || post) {
      Z = 1.0 / (double)ntot;
      S = 0.0;

      for (i = 0; i < mlist->nmodels; i++) {
	*(score+i) *= Z;
	if (post)
	  S += exp(*(score+i));
      }
      
      if (post) {
	Z = 1.0 / S;
	for (i = 0; i < mlist->nmodels; i++)
	  *(score+i) = exp(*(score+i)) * Z;
      }
    }
    /* output format is:  filename start_time nframes name nmodels llk[0] .... llk[n] */
    fprintf(ostream, "%s %.2f %.2f %ld", ifn, ASEG_NULL_TIME, ASEG_NULL_TIME, ntot);
    if (curseg)
      for (i = 0; i < curseg->label->nlabels; i++)
	fprintf(ostream, i ? "+%s" : " %s", curseg->label->name[i]);
    else
      fprintf(ostream, " -");
    fprintf(ostream, " %d", mlist->nmodels);
    for (i = 0; i < mlist->nmodels; i++)
      fprintf(ostream, " %-10.4f", *(score+i));
    fprintf(ostream, "\n");
  }

  /* reset memory */
  spf_stream_close(f);
  seg_list_free(seg);
  
  return(nsegs);
}

/* ------------------------------------------------------------------------------- */
/* ----- asseg_t *get_next_seg(float *, float *, asseg_t *, seglab_t *, int) ----- */
/* ------------------------------------------------------------------------------- */
/*
 * Get next segment start and end times. Return the current segment in
 * the input segmentation if the latter is specfied, NULL
 * otherwise. If a fixed segment length has been specified, return the
 * next (fixed) segment start and end times. In any other cases,
 * return the unique segment specified by the start and end times if
 * any.
 *
 * NOTE: in fixed length segment mode, st and et must be kept
 * untouched between two succesive calls without reset. When using an
 * input segmentation, the return pointer should be kept untouched
 * between two successive calls without reset since it must be
 * provided as the input to the next invocation of the function in
 * order to be able to move to the next segment.
 */
asseg_t *get_next_seg(float *st, float *et, asseg_t *seg, seglab_t *label, int reset)
{
  asseg_t *p = NULL;

  if (seg) { /* use input segmentation */
    p = (reset) ? seg : seg->next;
    if (! p) {
      *st = *et = ASEG_NULL_TIME;
      return(NULL);
    }
    while (seg_label_compare(p->label, label) != 0)
      if ((p = p->next) == NULL) {
	*st = *et = ASEG_NULL_TIME;
	return(NULL);
      }
    *st = p->st;
    *et = p->et;
  }
  else if (length > 0.0) { /* use fixed length segments */
    if (reset) {
      *st = startt;
      *et = length;
    }
    else {
      *st += shift;
      *et = (*st) + length;
      if (endt != ASEG_NULL_TIME && *st > endt)
	*st = *et = ASEG_NULL_TIME;
    }
  }
  else { /* use global start and end times */
    if (reset) {
      *st = startt;
      *et = endt;
    }
    else {
      *st = *et = ASEG_NULL_TIME;
    }
  }

  return(p);
}

/* ------------------------------------------------------------------------------------------- */
/* ----- unsigned long score_segment(spfstream_t *, float, float, gmmlist_t *, double *) ----- */
/* ------------------------------------------------------------------------------------------- */
/*
 * Score a given segment as specified by start and end times. If end
 * time is null, the segment extends until the end of the input
 * stream. Return the number of frames in the segment.
 */
unsigned long score_segment(spfstream_t *f, float st, float et, gmmlist_t *mlist, double *logp)
{
  unsigned short k;
  unsigned long t, sn, en, nframes;
  spf_t *yt;
  int idx[GMM_NUM_COMP_MAX];
  double a;

  if (st == ASEG_NULL_TIME)
    return(0);

  sn = (unsigned long)(st * spf_stream_rate(f));
  if (et != ASEG_NULL_TIME) {
    en = (unsigned long)(et * spf_stream_rate(f));
    nframes = (en - sn > 0) ? (en - sn) : 1;
  }
  else
    nframes = 0;
    
  /* try to seek to initial vector -- if not possible, it means the
     segment does not exist and we exit. */
  if (spf_stream_seek(f, sn, SEEK_SET) != 0)
    return(0);

  for (k = 0; k < mlist->nmodels; k++)
    *(logp+k) = 0.0;

  t = 0;

  while ((yt = get_next_spf_frame(f)) != NULL) {
   
    if (nbest == 0 || nbest >= get_gmm_num_comp(mlist->gmm[0])) {
      for (k = 0; k < mlist->nmodels; k++) {
	a = gmm_frame_log_like(yt, mlist->gmm[k]);
	*(logp+k) += a;
      }
    }
    else {
      a = gmm_nbest_frame_log_like(yt, mlist->gmm[0], idx, nbest, 1);
      *logp += a;

      for (k = 1; k < mlist->nmodels; k++) {
	a = gmm_nbest_frame_log_like(yt, mlist->gmm[k], idx, nbest, 0);
	*(logp+k) += a;
      }
    }

    t++;
    if (t == nframes)
      break;
  }

  return(t);
}

/* ------------------------------------------------------- */
/* ----- gmmlist_t *gmm_list_read(const char *, int) ----- */
/* ------------------------------------------------------- */
/*
 * Read a list of GMM specified in an input file. If mode == 0, a
 * single entry list is created where fn correspond to a single
 * model. Otherwise, model names are read from fn.  
 */
gmmlist_t *gmm_list_read(const char *fn, int mode)
{
  FILE *f;
  gmm_t **p = NULL;
  unsigned short n;
  gmmlist_t *list;
  char line[512];
 
  if ((list = (gmmlist_t *)malloc(sizeof(gmmlist_t *))) == NULL) {
    fprintf(stderr, "sgmlike error -- cannot allocate memory\n");
    return(NULL);
  }

  p = NULL;
  n = 0;

  if (mode == 0) {
    
    if ((p = (gmm_t **)malloc(sizeof(gmm_t *))) == NULL) {
      fprintf(stderr, "sgmlike error -- cannot allocate memory\n");
      return(NULL);
    }

    if ((*p = gmm_read(fn)) == NULL) {
      fprintf(stderr, "sgmlike error -- cannot read model from file %s\n", fn);
      free(p);
      return(NULL);
    }

    n++;
  }
  else {

    if ((f = fopen(fn, "r")) == NULL) {
      fprintf(stderr, "sgmlike error -- cannot open model list file %s\n", fn);
      free(list);
      return(NULL);
    }
      
    while (fgets(line, 512, f) != NULL) {

      if (line[strlen(line)-1] == '\n')
	line[strlen(line)-1] = 0;
      
      n++;
      
      if ((p = (gmm_t **)realloc(p, n * sizeof(gmm_t *))) == NULL) {
	fprintf(stderr, "sgmlike error -- cannot allocate memory\n");
	fclose(f);
	return(NULL);
      }
      
      if ((p[n-1] = gmm_read(line)) == NULL) {
	fprintf(stderr, "sgmlike error -- cannot read model from file %s\n", line);
	n--; while (n) free(p[--n]); free(p); fclose(f);
	return(NULL);
      }
    }

    fclose(f);
  }

  list->gmm = p;
  list->nmodels = n;

  return(list);
}

/* ------------------------------------------- */
/* ----- void gmm_list_free(gmmlist_t *) ----- */
/* ------------------------------------------- */
void gmm_list_free(gmmlist_t *p)
{
  unsigned short i;

  if (p) {
    for (i = 0; i < p->nmodels; i++)
      gmm_free(p->gmm[i]);   
    free(p);
  }
}

/* --------------------------------------- */
/* ----- int read_args(int, char **) ----- */
/* --------------------------------------- */
/*
 *  -s, --segmentation[=fn]      use segmentation for scoring (don't)
 *  -l, --label=s                score segments with specified name (none)
 *  -n, --length=n               segment length in seconds (all file)
 *  -d, --shift=n                segment shift in seconds (segment length)
 *  -t, --time=[st][:et]         score between start and end times (all file)
 *  -a, --add                    add segment scores (don't)
 *  -b, --n-best=n               N-best Gaussian scoring (no)
 *  -n, --normalize              normalize score by segment length (don't)
 *  -p, --posterior              output posterior probabilities (log-likelihoods)
 *  -L, --file-list=fn           input script (none)
 *  -f, --list=fn                set model list (none)
 *  -C, --convert=s              convertion string - ADNZ (none)
 *  -S, --input-bufsize=n        input buffer size in kbytes
 *  -o, --output=fn              output file (stdout)
 *  -v, --verbose                verbose mode
 *  -V, --version                print version number
 *  -h, --help                   help
 */
int read_args(int argc, char **argv)
{
  int c, i;
  char *p;
  struct option longopts[] = {
    {"segmentation", optional_argument, NULL, 's'},
    {"label", required_argument, NULL, 'x'},
    {"length", required_argument, NULL, 'l'},
    {"shift", required_argument, NULL, 'd'},
    {"time", required_argument, NULL, 't'},
    {"add", no_argument, NULL, 'a'},
    {"n-best", required_argument, NULL, 'b'},
    {"normalize", no_argument, NULL, 'n'},
    {"posterior", no_argument, NULL, 'p'},
    {"file-list", required_argument, NULL, 'L'},
    {"list", required_argument, NULL, 'f'},
    {"convert", required_argument, NULL, 'C'},
    {"buffer-size", required_argument, NULL, 'S'},
    {"output", required_argument, NULL, 'o'},
    {"verbose", optional_argument, NULL, 'v'},
    {"version", no_argument, NULL, 'V'},
    {"help", no_argument, NULL, 'h'},
    {0, 0, 0, 0}
  };

  opterr = 0;
  optopt = '!';

  while ((c = getopt_long(argc, argv, "s::x:l:d:t:ab:npL:f:C:S:o:v::Vh", longopts, &i)) != EOF)
    switch (c) {
      
    case 's':
      use_seg = 1;
      segfn = optarg;
      break;
      
    case 'x':
      labname = optarg;
      break;
      
    case 'l': 
      length = strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgmlike error -- invalid argument %s to --length (use --help to get usage)\n", optarg);
	return(1);
      }
      break;
      
    case 'd': 
      shift = strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgmlike -- invalid argument %s to --shift (use --help to get usage)\n", optarg);
	return(1);
      }
      break;
      
    case 't':
      p = optarg;
      if (*p != ':') {
	startt = (float)strtod(optarg, &p);
	if (p == optarg || (*p != 0 && *p != ':')) {
	  fprintf(stderr, "sgmlike error -- invalid argument %s to --time (use --help to get usage)\n", optarg);
	  return(1);
	}
      }
      if (*p == ':')
	endt = (float)strtod(p+1, &p);
      
      if (*p) {
	fprintf(stderr, "sgmlike error -- invalid argument %s to --time (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'a':
      add = 1;
      break;

    case 'b':
      nbest = (unsigned short)strtol(optarg, &p, 0);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgmlike error -- invalid argument %s to --n-best (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'n':
      normalize = 1;
      break;
      
    case 'p':
      post = 1;
      break;
      
    case 'L':
      scpfn = optarg;
      break;
      
    case 'f':
      listfn = optarg;
      break;
      
    case 'C': 
      cvtflag = sp_str_to_flag(optarg);      
      break;
      
    case 'S':
      iobs = (size_t)strtol(optarg, &p, 0) * 1000;
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sgmlike error -- invalid argument %s to --buffer-size (use --help to get usage)\n", optarg);
	return(1);
      }
      break;
      
    case 'o':
      ofn = optarg;
      break;
      
    case 'v':
      trace = 1;
      if (optarg) {
        trace = (int)strtol(optarg, &p, 0);
        if (p == optarg || *p != 0) {
          fprintf(stderr, "sgmlike error -- invalid argument %s to --verbose (use --help to get usage)\n", optarg);
          return(1);
        }
      }
      break;
      
    case 'V':
      version = 1;
      break;
      
    case 'h':
      help = 1;
      return(0);
      break;
      
    default:
      fprintf(stderr, "sgmlike error -- unrecognized option %s (use --help to get usage)\n", argv[optind-1]);
      return(1);
    }
  
  /* check argument integrity */
  if (shift == 0)
    shift = length;

  if (use_seg == 1 && length != 0)
    fprintf(stderr, ">>>>> WARNING >>>>> both segmentation and fixed-length segments specified (using segmentation)\n");

  if (use_seg == 1 && (startt != 0 || endt != ASEG_NULL_TIME))
    fprintf(stderr, ">>>>> WARNING >>>>> both segmentation and time constraint specified (using segmentation)\n");

  if (use_seg == 1 && scpfn == NULL && segfn == NULL) {
    fprintf(stderr, "sgmlike error -- cannot specify --segmentation without argument if no script is specified (use --help to get usage)\n");
    return(1);
  }

  return(0);
}

#undef _sgmlike_c_
