/******************************************************************************/
/*                                                                            */
/*                                sviterbi.c                                  */
/*                                                                            */
/*                        Audio Segmentation Library                          */
/*                                                                            */
/* Mike                                                             Aug, 2002 */
/* -------------------------------------------------------------------------- */
/*  Copyright (C) 2002 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * CVS log:
 *
 * $Author: ggravier $
 * $Date: 2003/11/20 15:50:58 $
 * $Revision: 1.11 $
 *
 */

/*
 * Viterbi decoding with GMM-based state densities.
 *
 * The HMM network is specified in an ASCII file according to the following
 * syntax:
 *
 *  nstates
 *  state_name /path/to/the/file.gmm
 *  ...
 *  a(I,1) a(I,2) ... a(I,N)
 *  a(1,1)     ...    a(1,N)
 *             ...
 *  a(N,1)     ...    a(N,N)
 *  a(1,F)     ...    a(N,F)
 *
 * where, in the transition matrix, N is the number of states, I and F are
 * the virtual initial and final states respectively.
 */

#define _sviterbi_c_

#include <audioseg.h>
#include <hmm.h>
#include <getopt.h>

static char *cvsid = "$Id: sviterbi.c,v 1.11 2003/11/20 15:50:58 ggravier Exp $";


/* ----------------------------------------------- */
/* ----- global variables set by read_args() ----- */
/* ----------------------------------------------- */
float penalty = 0.0;           /* inter-model penalty                  */
float akaike = 0.0;            /* penalize likelihood by complexity    */
float st = ASEG_NULL_TIME;     /* start tim                           */
float et = ASEG_NULL_TIME;     /* end time                             */
long cvtflag = 0;              /* qualifier convertion string          */
size_t iobs = 10000000;        /* I/O buffer size (in bytes)           */
int help = 0;                  /* print usage                          */
int trace = 0;                 /* trace level                          */
int version = 0;               /* version flag                         */

/* ---------------------------- */
/* ----- void usage(void) ----- */
/* ---------------------------- */
void usage(void)
{
  fprintf(stdout, "Usage:\n"); 
  fprintf(stdout, "  sviterbi [options] hmm data ofile\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Synopsis:\n");
  fprintf(stdout, "  Basic Viterbi decoder.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Options:\n");
  fprintf(stdout, "  -a, --akaike=f               complexity penality\n");
  fprintf(stdout, "  -p, --penalty=f              inter-class penalty\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -t, --time=[st][:et]         specifiy start and end times (whole file)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "  -C, --convert=s              convertion string - ZRADN (none)\n");
  fprintf(stdout, "  -S, --buffer-size=n          input buffer size in kbytes (%d)\n", iobs / 1000);  
  fprintf(stdout, "\n");
  fprintf(stdout, "  -v, --verbose[=n]            verbose mode\n");
  fprintf(stdout, "  -V, --version                print version number\n");
  fprintf(stdout, "  -h, --help                   this help message\n");
  fprintf(stdout, "\n");
}

/* --------------------------------- */
/* ----- int main(int,char **) ----- */
/* --------------------------------- */
int main(int argc,char **argv)
{
  char *hmmfn;
  char *datfn;
  char *segfn;

  hmm_t *hmm;
  spfstream_t *f;
  asseg_t *seg;

  int read_args(int, char **);
  asseg_t *decode(hmm_t *, spfstream_t *);

  /* ----- process command line ----- */
  if (read_args(argc, argv))
    return(1);

  if (help) {
    usage();
    return(0);
  }

  if (version) {
    fprintf(stdout, "audioseg %s -- %s\n", VERSION, cvsid);
    return(0);
  }

  /* read arguments */
  if (optind < argc)
    hmmfn = argv[optind++];
  else {
    fprintf(stderr, "sviterbi error -- no model filename specified (use --help to get usage)\n");
    return(1);
  }

  if (optind < argc)
    datfn = argv[optind++];
  else {
    fprintf(stderr, "sviterbi error -- no data filename specified (use --help to get usage)\n");
    return(1);
  }
  
  if (optind < argc)
    segfn = argv[optind++];
  else {
    fprintf(stderr, "sviterbi error -- no output filename specified (use --help to get usage)\n");
    return(1);
  }

  if (optind != argc) {
    fprintf(stderr, "sviterbi error -- invalid number of arguments (use --help to ger usage)\n");
    return(1);
  }

  /* ---- reading hmm file and data file ---- */
  if ((hmm = hmm_read(hmmfn)) == NULL) {
    fprintf(stderr, "sviterbi error -- cannot read hmm file %s\n", hmmfn);
    return(1);
  }

  if ((f = spf_input_stream_open(datfn, cvtflag, iobs)) == NULL) {
    fprintf(stderr, "sviterbi error -- cannot open feature stream %s\n", datfn ? datfn : "stdin");
    hmm_free(hmm);
    return(1);
  }

  if (st != ASEG_NULL_TIME)
    if (spf_stream_seek(f, (unsigned long)(st * spf_stream_rate(f)), SEEK_SET) != 0) {
      fprintf(stderr, "sviterbi error -- cannot seek to t=%.2f in stream %s\n", st, datfn ? datfn : "stdin");
      return(1);
    }

  /* ---- testing validity of data ---- */ 
  if(spf_stream_dim(f) != get_gmm_dim(hmm->state[0].pdf)) {
    fprintf(stderr, "sviterbi error -- data size and gmm size do not match\n");
    hmm_free(hmm); spf_stream_close(f);
    return(1);
  }

  /* ---- decoding ---- */
  if (trace) {
    fprintf(stdout, "decoding file %s\n", datfn);
    fflush(stdout);
  }
  if ((seg = decode(hmm, f)) == NULL) {
    fprintf(stderr, "sviterbi error -- unable to decode file %s\n", datfn);
    hmm_free(hmm); spf_stream_close(f);
    return(1);
  }

  /* ---- write output segmentation ---- */
  if (trace) {
    fprintf(stdout, "writing segmentation to file %s\n", segfn);
    fflush(stdout);
  }
  if (seg_write(seg, segfn, "lsep") == 0) {
    fprintf(stderr, "sviterbi error -- unable to write segment list to file %s\n", segfn);
    hmm_free(hmm); spf_stream_close(f); seg_list_free(seg);
    return(1);
  }

  /* ---- the end ---- */
  hmm_free(hmm);
  spf_stream_close(f);
  seg_list_free(seg);

  return(0);
}

# define VSIZE 1000

typedef struct {
  double logp;
  int prev;
} varray_t;

/* ---------------------------------------------------- */
/* -----  asseg_t *decode(hmm_t *, spfstream_t *) ----- */
/* ---------------------------------------------------- */
/*
 * Viterbi decoder.
 */
asseg_t *decode(hmm_t *hmm, spfstream_t *f) 
{
  unsigned long n, t, en, nt;
  unsigned short nstates = hmm->nstates;

  spf_t *yt;
  varray_t **vtab, *p, *pp;
  state_t *statep;

  double a, sbest, logp;
  unsigned short i, j, jbest, b = 0;

  asseg_t *seg, *curseg;

  en = 0;
  if (et != ASEG_NULL_TIME) {
    en = (unsigned long)(et * spf_stream_rate(f));
    if (st != ASEG_NULL_TIME)
      en -= (unsigned long)(st * spf_stream_rate(f));
  }

  /* allocate Viterbi array */
  if ((vtab = (varray_t **)malloc(VSIZE * sizeof(varray_t *))) == NULL)
    return(NULL);

  nt = VSIZE;

  t = 0;

  while ((yt = get_next_spf_frame(f)) != NULL) {
    
    /* allocate array for current time */
    if (t == nt) {
      if ((vtab = (varray_t **)realloc(vtab, (nt + VSIZE) * sizeof(varray_t *))) == NULL) {
	while (t) free(vtab[--t]); free(vtab);
	return(NULL);
      }
      nt += VSIZE;
    }
    if ((vtab[t] = (varray_t *)malloc(nstates * sizeof(varray_t))) == NULL) {
      while (t) free(vtab[--t]); free(vtab);
      return(NULL);
    }

    p = vtab[t];
    sbest = -DBL_MAX;
    jbest = 0;
    
    /* fill-in current column */
    if (t == 0) {

      /* fprintf(stderr, "t=%ld", (long)0); */
      
      /* use init probs */
      for (i = 0; i < nstates; i++) {
	logp = hmm->entryp[i] + gmm_frame_log_like(yt, hmm->state[i].pdf);
	if (akaike)
	  logp -= akaike * hmm->state[i].complexity;

	(p+i)->logp = logp;
	(p+i)->prev = HMM_NULL_STATE;

	if (logp > sbest) {
	  sbest = logp;
	  jbest = i;
	}
	
	/* fprintf(stderr, " %12.7f[-]", logp); */
      }
      /* fprintf(stderr, "\n"); */
    }
    else {
      /* fprintf(stderr, "t=%ld", t); */
      
      pp = vtab[t-1];

      for (i = 0; i < nstates; i++) {
	statep = hmm->state + i;

	logp = -DBL_MAX;
	for (j = 0; j < nstates; j++) {
	  a = (pp+j)->logp + hmm->transp[j][i];
	  if (i != j)
	    a -= penalty;
	  if (a > logp) {
	    logp = a;
	    b = j;
	  }
	}
    	logp += gmm_frame_log_like(yt, statep->pdf);
	if (akaike)
	  logp -= akaike * statep->complexity;

	(p+i)->logp = logp;
	(p+i)->prev = b;

	if (logp > sbest) {
	  sbest = logp;
	  jbest = i;
	}

	/* fprintf(stderr, " %12.7f[%d]", logp, b); */
      }
      
      /* fprintf(stderr, "\n"); */
    }

    if (trace > 1) {
      fprintf(stdout, "t=%5ld   label=%20s   avg. logp=%.2e\n", t, hmm->state[jbest].name, sbest);
      fflush(stdout);
    }

    t++;

    if (en > 0 && t == en)
      break;
  }

  /* find out best exit state */
  n = t;
  p = vtab[n-1];
  logp = -FLT_MAX;;
  for (i = 0; i < nstates; i++) {
    a = (p+i)->logp + hmm->exitp[i];

    if (a > logp) {
      logp = a;
      b = i;
    }
  }

  /* backtrack */
  statep = hmm->state + b;

  if ((seg = seg_create(statep->name, 0, t / spf_stream_rate(f), (p+b)->logp)) == NULL) {
    while (t) free(vtab[--t]); free(vtab);
    return(NULL);
  }

  /* fprintf(stderr, "t=%ld best=%d end_time=%.2f score=%.7f\n", t, b, t / spf_stream_rate(f), (p+b)->logp); */

  for (t = n - 1; t > 1; t--) {
    if (vtab[t][b].prev != b) { /* segment boundary found between frames t-1 and t */
      b = vtab[t][b].prev;
      p = vtab[t-1];
      a = t / spf_stream_rate(f);
      statep = hmm->state + b;

      if ((curseg = seg_create(statep->name, 0, a, (p+b)->logp)) == NULL) {
	seg_list_free(seg); t = n; while (t) free(vtab[--t]); free(vtab);
	return(NULL);
      }

      /* fprintf(stderr, "t=%ld best=%d end_time=%.2f score=%.7f diff=%.7f\n", t, b, a, (p+b)->logp, seg->score - (p+b)->logp); */

      seg->st = a;
      seg->score -= (p+b)->logp;

      curseg->next = seg;
      seg = curseg;
    }
  }
  
  /* add offset to segment boundaries if st was specified */
  if (st != ASEG_NULL_TIME) {
    a = (unsigned long)(st * spf_stream_rate(f)) / spf_stream_rate(f);
    curseg = seg;
    
    while (curseg) {
      curseg->st += a;
      curseg->et += a;
      curseg = curseg->next;
    }
  }

  return(seg);
}

/* --------------------------------------- */
/* ----- int read_args(int, char **) ----- */
/* --------------------------------------- */
/*
 *  -a, --akaike=f               complexity penality
 *  -p, --penalty=f              inter-class penalty
 *  -s, --start=f                start decoding at time
 *  -e, --end=f                  end decoding at time
 *  -C, --convert=s              convertion string - ADNZ (none)
 *  -v, --verbose               verbose mode
 *  -V, --version               print version number
 *  -h, --help
 */
int read_args(int argc, char **argv)
{
  int c, i;
  char *p;
  struct option longopts[] = {
    {"penalty", required_argument, NULL, 'p'},  
    {"akaike", required_argument, NULL, 'a'},
    {"time", required_argument, NULL, 't'},
    {"convert", required_argument, NULL, 'C'},
    {"buffer-size", required_argument, NULL, 'S'},
    {"verbose", no_argument, NULL, 'v'},
    {"version", no_argument, NULL, 'V'},
    {"help", no_argument, NULL, 'h'},
    {0, 0, 0, 0}
  };

  opterr = 0;
  optopt = '!';

  while ((c = getopt_long(argc, argv, "p:a:t:C:S:vVh", longopts, &i)) != EOF)
    switch (c) {

    case 'p':
      penalty = (float)strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sviterbi error -- invalid argument %s to --penalty (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'a':
      akaike = (float)strtod(optarg, &p);
      if (p == optarg || *p != 0) {
	fprintf(stderr, "sviterbi error -- invalid argument %s to --akaike (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 't':
      p = optarg;
      if (*p != ':') {
	st = (float)strtod(optarg, &p);
	if (p == optarg || (*p != 0 && *p != ':')) {
	  fprintf(stderr, "sviterbi error -- invalid argument %s to --time (use --help to get usage)\n", optarg);
	  return(1);
	}
      }
      if (*p == ':')
	et = (float)strtod(p+1, &p);
      
      if (*p) {
	fprintf(stderr, "sviterbi error -- invalid argument %s to --time (use --help to get usage)\n", optarg);
	return(1);
      }
      break;

    case 'C': 
      cvtflag = sp_str_to_flag(optarg);      
      break;

    case 'S':
      iobs = (size_t)strtol(optarg, &p, 0) * 1000;
      if (p == optarg || *p != 0) {
        fprintf(stderr, "sviterbi error -- invalid argument %s to --buffer-size (use --help to get usage)\n", optarg);
        return(1);
      }
      break;
      
    case 'v':
      trace = 1;
      if (optarg) {
	trace = strtol(optarg, &p, 0);
	if (p == optarg || *p != 0) {
	  fprintf(stderr, "sviterbi error -- invalid argument %s to --verbose\n", optarg);
	  return(1);
	}
      }
      break;

    case 'V':
      version = 1;
      return(0);
      break;

    case 'h':
      help = 1;
      return(0);
      break;

    default:
      fprintf(stderr, "sviterbi error -- unrecognized option %s (use --help to get usage)\n", argv[optind-1]);
      return(1);
    }

  return(0);
}

#undef _sviterbi_c_
